/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	event.h
 *	This header defines event used by mmcp functions.
 *
 *	Contact :	Jaesung Park (knucsid@gmail.com)
 *				Jiin Kim
 */


#ifndef	_MMCP_EVENT_H
#define	_MMCP_EVENT_H


/* ----------------------------------------------------------------- *
 *	Header Files Included
 * ----------------------------------------------------------------- */

#include <header/types.h>
#include <header/wrapper.h>

#include <sys/time.h>
#include <stdlib.h>


#ifdef	__cplusplus
extern "C" {
#endif

/* ----------------------------------------------------------------- *
 *	Definitions for MMCP Internal Event
 * ----------------------------------------------------------------- */

//	MMCP Internal Event Handlers (POSIX Thread functions)

/* event.c --------------------------------------------------------- */

extern void *mmcp_main_loop( void *arg );
extern void *mmcp_event_handler( void *arg );
extern void *mmcp_join_handler( void *arg );
extern void *mmcp_handover_handler( void *arg );
extern void *mmcp_capture_handler( void *arg );


#ifndef	_MMCP_DEBUG

#define	MMCP_MAINLOOP_TIMEOUT_SEC	0
#define	MMCP_MAINLOOP_TIMEOUT_USEC	100000	/* 100 [msec] */

#elif defined (_MMCP_DEBUG)

#define	MMCP_MAINLOOP_TIMEOUT_SEC	10		/* 1 [sec] */
#define	MMCP_MAINLOOP_TIMEOUT_USEC	0

#endif


//	MMCP Internal Event Handler Set.

struct mmcp_handleset
{
	//	USAGE:
	//		whlie( mso->hset.handler_is_running ) {
	//			// some event handler operations.
	//		}

	__u8	main_loop_is_running;
	__u8	event_handler_is_running;
	__u8	join_handler_is_running;
	__u8	handover_handler_is_running;
	__u8	capture_handler_is_running;

	pthread_t	loop;	/* Thread ID */
	void*		(*main_loop)		(void *arg);
	pthread_t	event;
	void*		(*event_handler)	(void *arg);

	pthread_t	join;
	void*		(*join_handler)		(void *arg);
	pthread_t	handover;
	void*		(*handover_handler)	(void *arg);

	pthread_t	capture;
	void*		(*capture_handler)	(void *arg);
};

/* event.c --------------------------------------------------------- */

extern void mmcp_handleset_init( struct mmcp_handleset *hset,
								 mmcp_role_t role );


/* ----------------------------------------------------------------- *
 *	Definitions for MMCP Internal Event Queue
 * ----------------------------------------------------------------- */

//	MMCP Events Description
//		Request event:	MMCP_REQ_XXX
//		Confirm event:	MMCP_CON_XXX
//		Deny	event:	MMCP_DNY_XXX
//		Expire	event:	MMCP_EXP_XXX
//		Execute event:	MMCP_EXE_XXX
//		ACK		event:	MMCP_ACK_XXX
//		SND		event:	MMCP_SND_XXX
//		RCV		event:	MMCP_RCV_XXX

typedef enum
{
	MMCP_REQ_SESSION_JOIN = 0,
	MMCP_CON_SESSION_JOIN,
	MMCP_DNY_SESSION_JOIN,
	MMCP_EXP_SESSION_JOIN,

	MMCP_REQ_LOCAL_JOIN,
	MMCP_CON_LOCAL_JOIN,
	MMCP_DNY_LOCAL_JOIN,
	MMCP_EXP_LOCAL_JOIN,

	MMCP_REQ_USER_LEAVE,
	MMCP_CON_USER_LEAVE,
	MMCP_DNY_USER_LEAVE,
	MMCP_EXP_USER_LEAVE,

	MMCP_SND_STATUS_REPORT,
	MMCP_RCV_STATUS_REPORT,

	MMCP_SND_ASTATUS_REPORT,
	MMCP_RCV_ASTATUS_REPORT,

	MMCP_SND_STATUS_PROBE,
	MMCP_RCV_STATUS_PROBE,
	MMCP_EXP_STATUS_PROBE,

	MMCP_REQ_HANDOVER_INIT,
	MMCP_CON_HANDOVER_INIT,
	MMCP_DNY_HANDOVER_INIT,
	MMCP_EXP_HANDOVER_INIT,

	MMCP_EXE_HANDOVER_TRANSFER,
	MMCP_ACK_HANDOVER_TRANSFER,
	MMCP_EXP_HANDOVER_TRANSFER

} mmcp_event_t;

struct mmcp_eventitem {
	struct mmcp_eventitem *prev;
	struct mmcp_eventitem *next;
	struct mmcp_eventq *q;

	mmcp_role_t role;
	mmcp_event_t type;
	mmcp_addr_t from;

	__u32	packetlen;
	void	*packet;
};

#define	MMCP_EVENTITEM_LEN	sizeof(struct mmcp_eventitem)

/* event.c --------------------------------------------------------- */

extern void mmcp_event_type_print( mmcp_event_t type );

extern struct mmcp_eventitem *mmcp_create_event( mmcp_event_t type, 
				mmcp_role_t role, mmcp_addr_t *from, size_t fromlen, 
				void *packet, __u32 packetlen );

extern void mmcp_destroy_event( struct mmcp_eventitem *e );

static inline struct mmcp_eventitem *mmcp_ei_alloc()
{
	struct mmcp_eventitem *_new = NULL;

	_new = (struct mmcp_eventitem *) calloc(1, MMCP_EVENTITEM_LEN);
	if( !_new )
		err_ret( "mmcp_ei_alloc() : calloc() failed.", NULL );

	return _new;
}

static inline void mmcp_ei_free( struct mmcp_eventitem *ei )
{
	if( ei->packetlen )
		free(ei->packet);

	if( ei )
		free(ei);

	ei = NULL;
}

static inline void mmcp_ei_add( struct mmcp_eventitem *_new,
								struct mmcp_eventitem *prev,
								struct mmcp_eventitem *next )
{
	next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

static inline void mmcp_ei_del( struct mmcp_eventitem *entry )
{
	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}

	
//	MMCP Evnet Queue Structure
struct mmcp_eventq {

	struct mmcp_eventitem *head;
	struct mmcp_eventitem *tail;

	mmcp_mutex_t mutex;

	mmcp_sem_t sem;

	__u32 num;
};

/* event.c --------------------------------------------------------- */

extern int mmcp_event_initq( struct mmcp_eventq *q );
extern int mmcp_event_enq( struct mmcp_eventq *q, struct mmcp_eventitem *_new );
extern struct mmcp_eventitem *mmcp_event_deq( struct mmcp_eventq *q );

static inline int mmcp_event_emptyq( struct mmcp_eventq *q )
{
	return (q->num == 0);
}


/* ----------------------------------------------------------------- *
 *	Definitions for MMCP Timer Event (Alarm)
 * ----------------------------------------------------------------- */

//	void mmcp_timeradd( struct timeval *a,
//						struct timeval *b,
//						struct timeval *result )
#define	mmcp_timeradd(a, b, result)						\
do {													\
	(result)->tv_sec  = (a)->tv_sec + (b)->tv_sec;		\
	(result)->tv_usec = (a)->tv_usec + (b)->tv_usec;	\
	if( (result)->tv_usec >= 1000000 ) {				\
		(result)->tv_sec++;								\
		(result)->tv_usec -= 1000000;					\
	}													\
} while( 0 )

//	void mmcp_timersub( struct timeval *a,
//						struct timeval *b,
//						struct timeval *result )
#define	mmcp_timersub(a, b, result)						\
do {													\
	(result)->tv_sec  = (a)->tv_sec - (b)->tv_sec;		\
	(result)->tv_usec = (a)->tv_usec - (b)->tv_usec;	\
	if( (result)->tv_usec < 0 ) {				\
		(result)->tv_sec--;								\
		(result)->tv_usec += 1000000;					\
	}													\
} while( 0 )

struct mmcp_timeritem
{
	struct mmcp_timeritem	*prev;
	struct mmcp_timeritem	*next;
	struct mmcp_timerlist	*list;

	__u32					id;				/* Timer ID */
	mmcp_timer_t			type;			/* Timer Type */
	__u32					update_time;	/* Seconds for update timer */
	struct timeval			action_time;	/* Start Time + Timeout value */
	struct mmcp_skbuff		*skb;			// no implementation
};

#define MMCP_TIMERITEM_LEN		sizeof(struct mmcp_timeritem)

/* timer.c --------------------------------------------------------- */

static inline struct mmcp_timeritem *mmcp_ti_alloc()
{
	struct mmcp_timeritem *_new = NULL;

	_new = (struct mmcp_timeritem *) calloc(1, MMCP_TIMERITEM_LEN);
	if( !_new )
		err_ret( "mmcp_ti_alloc() : calloc() failed.", NULL );

	return _new;
}

extern void mmcp_ti_free( struct mmcp_timeritem *alarm );

static inline void mmcp_ti_add( struct mmcp_timeritem *_new,
								struct mmcp_timeritem *prev,
								struct mmcp_timeritem *next )
{
	next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

static inline int mmcp_ti_update( struct mmcp_timeritem *alarm )
{
	struct timeval ct, inc;

	if( !alarm )
		err_ret( "mmcp_ti_update() : given timer is NULL.", -1 );

	if( gettimeofday(&ct, (struct timezone *) NULL) < 0 )
		err_ret( "mmcp_ti_update() : gettimeofday() failed.", -1 );

	inc.tv_sec = alarm->update_time;
	inc.tv_usec = 0;

	mmcp_timeradd( &ct, &inc, &alarm->action_time );

	return 0;
}

static inline void mmcp_ti_unlink( struct mmcp_timeritem *entry )
{
	if( !entry )
		return;

	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}

static inline struct mmcp_timeritem *mmcp_create_timer( mmcp_timer_t type,
			int time, struct mmcp_skbuff *skb )
{
	struct mmcp_timeritem *_new = NULL;

	if( !(_new = mmcp_ti_alloc()) )
		err_ret( "mmcp_create_timer() : mmcp_ti_alloc() failed.", NULL );

	_new->type = type;
	_new->update_time = time;

	/* KNUCSID : check */
	_new->skb = skb;

	if( mmcp_ti_update(_new) < 0 ) {
		mmcp_ti_free(_new);
		_new = NULL;
		err_ret( "mmcp_create_timer() : mmcp_ti_update() failed.", NULL );
	}

	return _new;
}

static inline void mmcp_destroy_timer( struct mmcp_timeritem *alarm )
{
	mmcp_ti_free(alarm);
}

struct mmcp_timerlist
{
	struct mmcp_timeritem *head;
	struct mmcp_timeritem *tail;

	mmcp_mutex_t mutex;
	__u32 num;	
};

/* timer.c --------------------------------------------------------- */

extern int mmcp_timerlist_init( struct mmcp_timerlist *list );

static inline int mmcp_timerlist_empty( struct mmcp_timerlist *list )
{
	return (list->num == 0);
}

extern int mmcp_timerlist_add( struct mmcp_timerlist *list,
					struct mmcp_timeritem *_new );	// sorted add

extern struct mmcp_timeritem *mmcp_timerlist_del( struct mmcp_timerlist *list,
					struct mmcp_timeritem *alarm );

static inline struct mmcp_timeritem *mmcp_timerlist_peek(
										struct mmcp_timerlist *list )
{
	return list->head;
}

static inline struct timeval *mmcp_timerlist_peek_timeout(
										struct mmcp_timerlist *list )
{
	static struct timeval result;
	struct timeval current;

	if( mmcp_timerlist_empty(list) )
		return NULL;

	if( gettimeofday(&current, (struct timezone *) NULL) < 0 )
		err_ret( "mmcp_timerlist_peek_timeout() : "
				 "gettimeofday() failed.", NULL );

	mmcp_timersub( &list->head->action_time, &current, &result );
	
	if( result.tv_sec < 0 ) {
		result.tv_sec = 0;
		result.tv_usec = 0;
	}

	return &result;
}


#ifdef	__cplusplus
}
#endif

#endif	// _MMCP_EVENT_H
