/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	structs.h
 *	This header represents the MMCP socket structure 
 * 	and protocol control block.
 *
 *	Contact :	Jaesung Park (knucsid@gmail.com)
 *				Jiin Kim
 */


#ifndef	_MMCP_STRUCTS_H
#define	_MMCP_STRUCTS_H

/* ----------------------------------------------------------------- *
 *	Header Files Included
 * ----------------------------------------------------------------- */

#include <header/types.h>
#include <header/event.h>

#include <string.h>
#include <stdlib.h>


#ifdef	__cplusplus
extern "C" {
#endif


//#define	MMCP_MNID_LEN		16
//#define	MMCP_POAID_LEN		16


/* ----------------------------------------------------------------- *
 *	Forward Declarations for Data Structures
 * ----------------------------------------------------------------- */

struct mmcp_globals;
struct mmcp_socket;
struct mmcp_endpoint;
struct mmcp_sm;
struct mmcp_lmc;
struct mmcp_mn;
struct mmcp_skbuff;
struct mmcp_skblist;


/* ----------------------------------------------------------------- *
 *	MMCP Socket Buffer Structure
 * ----------------------------------------------------------------- */

/* The Maximum Number of The MMCP Socket */
#define	MMCP_MAX_SOCKETS	8

struct mmcp_globals {

	/* The Current Number of The MMCP Socket */
	__u8				nsock;

	/* Mutex for MMCP Socket */
	mmcp_mutex_t		mutex;

	/* The Global MMCP Socket Array Pointer */
	struct mmcp_socket	*socks[MMCP_MAX_SOCKETS];

};

/* MMCP Global Variable */
extern struct mmcp_globals *mmcp_global;

/* Socket Point that is used to Current Process. */
extern struct mmcp_socket *mmcp_current;

/* socket.c -------------------------------------------------------- */

extern struct mmcp_globals *mmcp_create_globals();

extern void mmcp_destroy_globals( struct mmcp_globals *global );

static inline struct mmcp_socket *mmcp_get_socket( int msd )
{
	return (mmcp_current = mmcp_global->socks[msd]);
}


/* ----------------------------------------------------------------- *
 *	MMCP Socket Buffer Structure
 * ----------------------------------------------------------------- */

struct mmcp_skbuff {
	struct mmcp_skbuff *prev;
	struct mmcp_skbuff *next;

	struct mmcp_skblist *list;

	__u32		sd;
	mmcp_addr_t addr;
	__u32		addrlen;
	__u32		pktlen;
	void		*pkt;
	__u16		has_sent;
	__u16		rtxcnt;
};

#define MMCP_SKBUFF_LEN		sizeof(struct mmcp_skbuff)

static inline struct mmcp_skbuff *mmcp_skb_alloc(
		struct mmcp_socket *mso, int sd, mmcp_addr_t *addr, socklen_t addrlen )
{
	struct mmcp_skbuff *skb = NULL;

	skb = (struct mmcp_skbuff *) malloc(MMCP_SKBUFF_LEN);
	if( !skb ) {
		return NULL;
	}

	skb->sd = sd;
	skb->addrlen = addrlen;
	memcpy( &skb->addr, addr, skb->addrlen );

	return skb;
}

static inline void mmcp_skb_free( struct mmcp_skbuff *skb )
{
	/* Check */
	if( skb->pktlen )
		free(skb->pkt);

	if( skb ) {
		free(skb);
		skb = NULL;
	}
}

static inline void mmcp_skb_add( struct mmcp_skbuff *_new,
								 struct mmcp_skbuff *prev,
								 struct mmcp_skbuff *next )
{
	next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

static inline void mmcp_skb_unlink( struct mmcp_skbuff *entry )
{
	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}


/* ----------------------------------------------------------------- *
 *	MMCP Socket Buffer List Structure
 * ----------------------------------------------------------------- */

struct mmcp_skblist {
	struct mmcp_skbuff *head;
	struct mmcp_skbuff *tail;

	mmcp_mutex_t mutex;
	mmcp_sem_t sem;

	__u32 num;
	__u32 length;
	__u32 max_length;
};

/* socket.c -------------------------------------------------------- */

extern int mmcp_skblist_init( struct mmcp_skblist *list, __u32 max_length );

static inline __u32 mmcp_skblist_spare( struct mmcp_skblist *list )
{
	return (list->max_length - list->length);
}

static inline __u32 mmcp_skblist_empty( struct mmcp_skblist *list )
{
	return (list->num == 0);
}

extern int mmcp_skblist_add( struct mmcp_skblist *list,
								struct mmcp_skbuff *_new );

extern struct mmcp_skbuff *mmcp_skblist_del( struct mmcp_skblist *list,
											 struct mmcp_skbuff *skb );


/* ----------------------------------------------------------------- *
 *	MMCP Socket Structure
 * ----------------------------------------------------------------- */

struct mmcp_socket {

	/* MMCP Control Socket Descriptor */
	__u32	msd;

	/* Multicast socket descriptor */
	__u32	usd;

	/* Domain (AF_INET or AF_INET6) */
	__u32	domain;

	/* Protocol (IPPROTO_UDP or IPPROTO_TCP) */
	__u32	protocol;

	/* Address */
	mmcp_addr_t addr;

	/* MMCP transport endpoint */
	struct mmcp_endpoint *ep;

	/* MMCP Timer List */
	struct mmcp_timerlist timer_list;

	/* MMCP Internal Event Handle Set */
	struct mmcp_handleset hset;

	/* MMCP Event Queue:
	 *	Join Queue		[SM/LMC/MN]
	 *	Handover Queue	[SM/LMC/MN]
	 *	Event Queue		[SM/LMC/MN]
	 */
	struct mmcp_eventq joinq, handoverq, eventq;

	/* Maximum length of the send buffer */
	__u32 max_sndbuf;

	/* Send buffers for Reliable Transport of Control Message */
	struct mmcp_skblist sndbuf;

};

#define MMCP_SOCKET_LEN		sizeof(struct mmcp_socket)

/* socket.c -------------------------------------------------------- */

extern struct mmcp_socket *mmcp_create_socket( int domain, int protocol, 
												mmcp_role_t role );
extern int mmcp_destroy_socket( struct mmcp_socket *mso );

extern struct mmcp_socket *mmcp_sk_alloc();
extern void mmcp_sk_free( struct mmcp_socket *mso );
extern int mmcp_sk_init( struct mmcp_socket *mso, int domain, int protocol,
							mmcp_role_t role );

extern int mmcp_sk_bind( struct mmcp_socket *mso, 
							struct sockaddr_storage *addr );


/* ----------------------------------------------------------------- *
 *	MMCP Endpoint Structure
 * ----------------------------------------------------------------- */

struct mmcp_endpoint {

	/* Backpointer to MMCP socket to which this endpoint belong. */
	struct mmcp_socket *mso;

	/* MMCP Finite State Machine */
	mmcp_state_t state;

	/* MMCP Role
	 *	MMCP_ROLE_SM	- Session Manager
	 *	MMCP_ROLE_LMC	- Local Mobility Controller
	 *	MMCP_ROLE_MN	- Mobile Node
	 */
	mmcp_role_t role;

	/* It depends on the role 
	 * If role is Session Manager, the mmcp_sm structure is enable
	 * If role is Local Mobility Controller, the mmcp_lmc structure is enable
	 * If role is Mobile Node, the mmcp_mn is enable
	 */
	struct mmcp_sm *sm;
	struct mmcp_lmc *lmc;
	struct mmcp_mn *mn;
};

#define MMCP_ENDPOINT_LEN		sizeof(struct mmcp_endpoint)

/* socket.c -------------------------------------------------------- */

extern int mmcp_create_endpoint( struct mmcp_socket *mso, mmcp_role_t role );
extern int mmcp_destroy_endpoint( struct mmcp_endpoint *ep );

extern struct mmcp_endpoint *mmcp_ep_alloc();
extern void mmcp_ep_free( struct mmcp_endpoint *ep );
extern int mmcp_ep_init( struct mmcp_socket *mso, struct mmcp_endpoint *ep );


/* ----------------------------------------------------------------- *
 *	MMCP Status Probe Timer Structure
 * ----------------------------------------------------------------- */

struct mmcp_spt_lmc {
	__s8	poa_id[MMCP_POAID_LEN];
	__u32	flag;
	__u32	count;
};

struct mmcp_spt_mn {
	__s8	mn_id[MMCP_MNID_LEN];
	__u32	flag;
	__u32	count;
};


/* ----------------------------------------------------------------- *
 *	MMCP Session Manager Structure
 * ----------------------------------------------------------------- */

struct mmcp_sm {

	struct mmcp_endpoint *sm_ep;

	mmcp_addr_t mcs_addr;
	mmcp_addr_t group_addr;

	mmcp_lmclist_t lmclist[4];	/* Temporary */
	mmcp_addr_t mnlist[8];		/* Temporary */
	int mn_index;

	struct mmcp_timeritem *spt;
	mn_status mmcp_mn_info[4][4];
	struct mmcp_spt_lmc spt_lmc[4];
};

#define MMCP_SM_LEN		sizeof(struct mmcp_sm)

/* socket.c -------------------------------------------------------- */

extern struct mmcp_sm *mmcp_sm_alloc();
extern void mmcp_sm_free( struct mmcp_sm *sm );
extern void mmcp_sm_init( struct mmcp_socket *mso, struct mmcp_sm *sm );


/* ----------------------------------------------------------------- *
 *	MMCP Local Mobility Controller Structure
 * ----------------------------------------------------------------- */

struct mmcp_lmc {

	struct mmcp_endpoint *lmc_ep;

	mmcp_addr_t sm_addr;
	mmcp_addr_t group_addr;

	mmcp_hd_t hd_val;	/* temporary value of mn information for handover */
	mmcp_addr_t mn_addr[4];		/* 4 is a temporary value */
	mmcp_lmclist_t lmc_list[3];		/* 3 is a temporary value */
	__s8	poa_id[16];

	struct mmcp_timeritem *sgt;
	struct mmcp_timeritem *spt;
	struct mmcp_spt_mn spt_mn[4];

	mn_status mmcp_mn_info[4];
};

#define MMCP_LMC_LEN	sizeof(struct mmcp_lmc)

/* socket.c -------------------------------------------------------- */

extern struct mmcp_lmc *mmcp_lmc_alloc();
extern void mmcp_lmc_free( struct mmcp_lmc *lmc );
extern void mmcp_lmc_init( struct mmcp_lmc *lmc );


/* ----------------------------------------------------------------- *
 *	MMCP Mobile Node Structure
 * ----------------------------------------------------------------- */

struct mmcp_mn {

	struct mmcp_endpoint *mn_ep;

	__s8 mn_id[MMCP_MNID_LEN];
	__s8 poa_id[MMCP_POAID_LEN];

	__u32	init_byte;
	__u32	init_packet;
	__s64	elapsed_time;
	
	mmcp_addr_t sm_addr;
	mmcp_addr_t lmc_addr;
	mmcp_addr_t newlmc_addr;
	mmcp_addr_t group_addr;

	struct mmcp_timeritem *sgt;
	char *ifname;
};

#define MMCP_MN_LEN		sizeof(struct mmcp_mn)

/* socket.c -------------------------------------------------------- */

extern struct mmcp_mn *mmcp_mn_alloc();
extern void mmcp_mn_free( struct mmcp_mn *mn );
extern void mmcp_mn_init( struct mmcp_mn *mn );

#ifdef	__cplusplus
}
#endif

#endif	/* _MMCP_STRUCTS_H */
