/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	types.h
 *	This header defines various types used by mmcp functions.
 *
 *	Contact :	Jaesung Park(knucsid@gmail.com)
 *				Jiin Kim
 */

#ifndef	_MMCP_TYPES_H
#define	_MMCP_TYPES_H

/* ----------------------------------------------------------------- *
 *	Header Files Included
 * ----------------------------------------------------------------- */

#include <header/wrapper.h>

#include <string.h>
#include <netinet/in.h>
#include <errno.h>
#include <pthread.h>	/* for POSIX thread */
#include <semaphore.h>	/* for POSIX semaphore */


#ifdef	__cplusplus
extern "C" {
#endif

#define	MMCP_MNID_LEN	16
#define MMCP_POAID_LEN	16

/* ----------------------------------------------------------------- *
 *	Definitions for primitive types
 * ----------------------------------------------------------------- */

typedef char			__s8;
typedef unsigned char	__u8;

typedef short			__s16;
typedef unsigned short	__u16;

typedef int				__s32;
typedef unsigned int	__u32;

#if defined(__GNUC__)
typedef long long			__s64;
typedef unsigned long long	__u64;
#endif


/* ----------------------------------------------------------------- *
 *	MMCP Role
 * ----------------------------------------------------------------- */

typedef enum
{
	MMCP_ROLE_NONE	= 0,
#define	MMCP_ROLE_NONE		MMCP_ROLE_NONE

	MMCP_ROLE_SM,
#define	MMCP_ROLE_SM		MMCP_ROLE_SM

	MMCP_ROLE_LMC,
#define	MMCP_ROLE_LMC		MMCP_ROLE_LMC

	MMCP_ROLE_MN
#define	MMCP_ROLE_MN		MMCP_ROLE_MN

} mmcp_role_t;


/* ----------------------------------------------------------------- *
 *	MMCP Timer
 * ----------------------------------------------------------------- */

typedef enum
{
	MMCP_TIMER_JWT	= 0,
	MMCP_TIMER_SRT,
	MMCP_TIMER_SPT,
	MMCP_TIMER_HIT,
	MMCP_TIMER_CTT,
	MMCP_TIMER_RXT
} mmcp_timer_t;

#define	MMCP_TIMER_JWT_VAL		5
#define	MMCP_TIMER_SRT_VAL		8
#define	MMCP_TIMER_SPT_VAL		18
#define	MMCP_TIMER_HIT_VAL		12
#define	MMCP_TIMER_CTT_VAL		9
#define	MMCP_TIMER_RXT_VAL		3

#define	MMCP_MAX_RXT			3


/* ----------------------------------------------------------------- *
 *	MMCP Address Type
 * ----------------------------------------------------------------- */

typedef union mmcp_sockaddr
{
	struct sockaddr_in		sa;
	struct sockaddr_in6		sa6;
} mmcp_addr_t;

#define MMCP_ADDR_LEN	sizeof(mmcp_addr_t)


typedef union mmcp_ip
{
	__u32		ip;
	char		ip6[16];
} mmcp_ip_t;

#define MMCP_IP_LEN		sizeof(mmcp_ip_t)


#define	SA			struct sockaddr
#define	SS			struct sockaddr_storage
#define	SA_IN		struct sockaddr_in
#define	SA_IN6		struct sockaddr_in6

#define	SA_LEN		sizeof(struct sockaddr)
#define	SS_LEN		sizeof(struct sockaddr_storage)
#define	SA_IN_LEN	sizeof(struct sockaddr_in)
#define	SA_IN6_LEN	sizeof(struct sockaddr_in6)

static inline socklen_t mmcp_addrlen( int domain )
{
	if( domain == AF_INET )
		return SA_IN_LEN;

	else if( domain == AF_INET6 )
		return SA_IN6_LEN;

	return 0;
}

static inline SA *mmcp_sockaddr( int domain, mmcp_addr_t *addr )
{
	if( domain == AF_INET )
		return (SA *) &addr->sa;

	else if( domain == AF_INET6 )
		return (SA *) &addr->sa6;

	return NULL;
}

static inline __u32 mmcp_addrchk( mmcp_addr_t *addr1, mmcp_addr_t *addr2 )
{
	__u32 result = -1;

	if( ((SS *) addr1)->ss_family != ((SS *) addr2)->ss_family )
		return -1;

	switch ( ((SS *) addr1)->ss_family ) {
		case AF_INET:
			result = memcmp(&((SA_IN *) addr1)->sin_addr,
					&((SA_IN *) addr2)->sin_addr, sizeof(struct in_addr));
			break;
		case AF_INET6:
			result = memcmp(&((SA_IN6 *) addr1)->sin6_addr,
					&((SA_IN6 *) addr2)->sin6_addr, sizeof(struct in6_addr));
			break;
		default:
			return -1;
	}

	return !result ? 0 : -1;
}


/* ----------------------------------------------------------------- *
 *	MMCP Interface Structure For Status Monitoring
 * ----------------------------------------------------------------- */

typedef struct _inter_io {
	__u32	byte;
//	__u32	packet;
//	__u32	error;
//	__u32	drop;
} inter_io;

typedef struct _interface_io {
	__s8	dev_name[16];
	inter_io	rcv_io;
} interface_io;

typedef struct _mn_status {
	__s8	mn_id[MMCP_MNID_LEN];
	__u32	rcv_byte;
	__u32	elapsed_time;
} mn_status;


/* ----------------------------------------------------------------- *
 *	Member Structure of MMCP Internal Structure
 * ----------------------------------------------------------------- */

typedef struct mmcp_lmclist {
	mmcp_addr_t lmc_addr;
	__s8		poa_id[MMCP_POAID_LEN];
} mmcp_lmclist_t;

typedef struct mmcp_hd_val {
	mmcp_addr_t hd_addr;		/* MN address (request handover) */
	mmcp_addr_t group_addr;		/* Group address of handover */
	mmcp_addr_t newlmc_addr;	/* New LMC address */
	__s8	mn_id[MMCP_MNID_LEN];
	__s8	poa_id[MMCP_POAID_LEN];
} mmcp_hd_t;


/* ----------------------------------------------------------------- *
 *	MMCP Finite State Machine
 * ----------------------------------------------------------------- */

typedef enum
{
	/* Common States */
	MMCP_STATE_CLOSED			= 0,
	MMCP_STATE_ESTABLISHED,

	// SM States
	MMCP_STATE_SJR_RECV,

	// LMC, MN States
	MMCP_STATE_SJR_SENT,
	MMCP_STATE_LJR_SENT,
	MMCP_STATE_HR_SENT,

} mmcp_state_t;


/* ----------------------------------------------------------------- *
 *	MMCP Default Protocol Parameters.
 * ----------------------------------------------------------------- */

#define	MMCP_PARAM_SNDBUF_DFL	(64 * 1024)
#define	MMCP_PARAM_RCVBUF_DFL	(64 * 1024)


/* ----------------------------------------------------------------- *
 *	Definitions for Synchronization.
 * ----------------------------------------------------------------- */

/* If we use POSIX Semaphore, ...
 *		int sem_init( sem_t *sem, int pshared, unsigned int value );
 *		int sem_wait( sem_t *sem );
 *		int sem_trywait( sem_t *sem );
 *		int sem_post( sem_t *sem );
 *		int sem_getvalue( sem_t *sem, int *sval );
 *		int sem_destroy( sem_t *sem );
 */

/* MMCP Semaphore Structure */
typedef struct mmcp_sem_struct
{
	sem_t	s;	/* POSIX Semaphore */
	__u32	n;	/* Currently available number of resources */
} mmcp_sem_t;

/* control.c ------------------------------------------------------- */

extern int mmcp_sem_init( mmcp_sem_t *sem, __u32 val );
extern int mmcp_sem_destroy( mmcp_sem_t *sem );

static inline int mmcp_sem_isempty( mmcp_sem_t *sem )
{
	return ( sem->n == 0 );
}

static inline int mmcp_sem_wait( mmcp_sem_t *sem )
{
	int result = -1;

	do {
		result = sem_wait(&sem->s);
	} while( result < 0 && errno == EINTR );

	if( result < 0 ) {
		return -1;
	}

	sem->n--;
	return 0;
}

static inline int mmcp_sem_trywait( mmcp_sem_t *sem )
{
	sem_trywait(&sem->s);
	sem->n--;
	return 0;
}

static inline int mmcp_sem_post( mmcp_sem_t *sem )
{
	if( sem_post(&sem->s) < 0 )
		err_ret( "mmcp_sem_post() : sem_post() failed.", -1 );
	sem->n++;

	return 0;
}

typedef mmcp_sem_t	mmcp_mutex_t;

static inline int mmcp_mutex_init( mmcp_mutex_t *mutex )
{
	return mmcp_sem_init( mutex, 1 );
}

static inline int mmcp_mutex_on( mmcp_mutex_t *mutex )
{
	return mmcp_sem_wait(mutex);
}

static inline int mmcp_mutex_off( mmcp_mutex_t *mutex )
{
	return mmcp_sem_post(mutex);
}

static inline int mmcp_mutex_destroy( mmcp_mutex_t *mutex )
{
	return mmcp_sem_destroy(mutex);
}


#ifdef	__cplusplus
}
#endif

#endif	// _MMCP_TYPES_H
