/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	control.c
 *	This file includes the implementation of the MMCP control.
 *
 *	Contact :	Jaesung Park(knucsid@gmail.com)
 *				Jiin Kim
 */

#include <header/mmcp.h>
#include <header/types.h>
#include <header/structs.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* ----------------------------------------------------------------- *
 *	Event Invocation Functions
 * ----------------------------------------------------------------- */

/* mmcp.h ---------------------------------------------------------- */

int mmcp_ops_start_events( struct mmcp_socket *mso )
{
	mmcp_role_t role = mso->ep->role;
	struct mmcp_handleset *hset = &mso->hset;

	hset->main_loop_is_running = 1;

	if( role == MMCP_ROLE_SM ) {

		hset->event_handler_is_running = 1;
		if( pthread_create(&hset->event, NULL, hset->event_handler, mso) ) {
			err_goto( "mmcp_ops_start_event(): "
					"cannot invoke event handler.\n", rb_start_event );
		}

		hset->join_handler_is_running = 1;
		if( pthread_create(&hset->join, NULL, hset->join_handler, mso) ) {
			err_goto( "mmcp_ops_start_event(): "
					"cannot invoke join handler.\n", rb_start_event );
		}

		hset->handover_handler_is_running = 1;
		if( pthread_create(&hset->handover, NULL, 
							hset->handover_handler, mso) ) {
			err_goto( "mmcp_ops_start_event(): "
					"cannot invoke handover handler.\n", rb_start_event );
		}

	}

	else if( role == MMCP_ROLE_LMC ) {
		if( pthread_create(&hset->loop, NULL, hset->main_loop, mso) ) {
			err_goto( "mmcp_ops_start_events(): "
					"cannot invoke main loop.", rb_start_event );
		}
	}

	else if( role == MMCP_ROLE_MN ) {
		if( pthread_create(&hset->loop, NULL, hset->main_loop, mso) ) {
			err_goto( "mmcp_ops_start_events(): "
					"cannot invoke main loop.", rb_start_event );
		}
	}

	else {
		err_goto( "mmcp_ops_start_events(): "
				"role does not exist.", rb_start_event );
	}
	return 0;

rb_start_event:	/* Roll back point */
	hset->main_loop_is_running			= 0;
	hset->event_handler_is_running		= 0;
	hset->join_handler_is_running		= 0;
	hset->handover_handler_is_running	= 0;

	return -1;
}

int mmcp_ops_end_events( struct mmcp_socket *mso )
{
	mmcp_role_t role = mso->ep->role;
	struct mmcp_handleset *hset = &mso->hset;

	hset->event_handler_is_running = 0;
	if( pthread_cancel(hset->event) != 0 ) {
		err_ret( "mmcp_ops_end_events(): "
				"cannot cancel event handler.", -1 );
	}

	if( role == MMCP_ROLE_SM || role == MMCP_ROLE_LMC ) {
		hset->join_handler_is_running = 0;
		if( pthread_cancel(hset->join) != 0 ) {
			err_ret( "mmcp_ops_end_events(): "
					"cannot cancel join handler.", -1 );
		}

		hset->handover_handler_is_running = 0;
		if( pthread_cancel(hset->handover) != 0 ) {
			err_ret( "mmcp_ops_end_events(): "
					"cannot cancel handover handler.", -1 );
		}
	}

	hset->main_loop_is_running = 0;
	if( pthread_cancel(hset->loop) != 0 ) {
		err_ret( "mmcp_ops_end_events(): "
				"cannot cancel main loop.", -1 );
	}

	return 0;
}

int mmcp_ops_maintain_events( struct mmcp_socket *mso )
{
	mmcp_role_t role = mso->ep->role;
	
	struct mmcp_handleset *hset = &mso->hset;
	int state;
	void *t_return;

	if( role == MMCP_ROLE_LMC ) {

		hset->join_handler_is_running = 1;
		if( pthread_create(&hset->join, NULL, hset->join_handler, mso) ) {
			hset->join_handler_is_running = 0;
			err_ret( "mmcp_ops_maintain_event(): "
					"cannot invoke join handler.\n", -1 );
		}

		hset->handover_handler_is_running = 1;
		if( pthread_create(&hset->handover, NULL, 
							hset->handover_handler, mso) ) {
			hset->handover_handler_is_running = 0;
			err_ret( "mmcp_ops_maintain_event(): "
					"cannot invoke handover handler.\n", -1 );
		}

		state = pthread_join(hset->loop, &t_return);
		if( state != 0 ) {
			err_ret( "mmcp_ops_maintain_event(): "
					"pthread_join() loop error\n", -1 );
		}

		state = pthread_join(hset->join, &t_return);
		if( state != 0 ) {
			err_ret( "mmcp_ops_maintain_event(): "
					"pthread_join() join error\n", -1 );
		}

		state = pthread_join(hset->handover, &t_return);
		if( state != 0 ) {
			err_ret( "mmcp_ops_maintain_event(): "
					"pthread_join() handover error\n", -1 );
		}
	}
	
	return 0;
}

/* ----------------------------------------------------------------- *
 *	State Machine Control Functions
 * ----------------------------------------------------------------- */

/* mmcp.h ---------------------------------------------------------- */

int mmcp_ops_start_join( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_lmc *lmc = NULL;
	struct mmcp_mn *mn = NULL;

	mmcp_pkt_sjr *sjr = NULL;
	mmcp_pkt_sjc *sjc = NULL;
	mmcp_pkt_ljr *ljr = NULL;
	mmcp_pkt_ljc *ljc = NULL;

	struct mmcp_skbuff *skb		= NULL;
	struct mmcp_timeritem *rxt	= NULL;
	struct mmcp_eventitem *e	= NULL;

	socklen_t addrlen;

//	char* str;	/* Temporary */

	/* Make Session Join Request Message. */
	if( !(sjr = mmcp_create_sjr(mso)) )
		err_ret( "mmcp_ops_start_join() : cannot create SJR.", -1 );

	/* Make Send Buffer Item. */
	addrlen = mmcp_addrlen(mso->domain);

	if( ep->role == MMCP_ROLE_LMC ) {
		lmc = mso->ep->lmc;
		skb = mmcp_skb_alloc( mso, mso->msd, &lmc->sm_addr, addrlen );
	}

	if( ep->role == MMCP_ROLE_MN ) {
		mn = mso->ep->mn;
		skb = mmcp_skb_alloc( mso, mso->msd, &mn->sm_addr, addrlen );
	}

	if( !skb )
		err_ret( "mmcp_ops_start_join() : mmcp_skb_alloc() failed.", -1 );

	skb->pktlen	= MMCP_PKT_SJR_LEN;
	skb->pkt	= sjr;

	mmcp_skblist_add( &mso->sndbuf, skb );

	/* Make Retransmission Timer (RXT) */
	rxt = mmcp_create_timer( MMCP_TIMER_RXT, MMCP_TIMER_RXT_VAL, skb );
	mmcp_timerlist_add( &mso->timer_list, rxt );

	mmcp_skb_transmit(skb);

	puts( "MMCP, Waiting Session Join Confirm! (STATE: SJR-SENT)" );
	ep->state = MMCP_STATE_SJR_SENT;

	while( 1 ) {

		e = mmcp_event_deq( &mso->joinq );
		
		/* Block: Waiting for SJC! */

		if( e->type == MMCP_CON_SESSION_JOIN ) {
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
			puts( "MMCP, Session Join Confirm!" );
			break;
		}

		else if( e->type == MMCP_EXP_SESSION_JOIN ) {
			/* RXT for SJR is already removed by mmcp_timer_dispatch(). */
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			err_ret( "MMCP, Session Join Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_start_join() : "
					 "Exceed Maximum Retransmission for SJR.", -1 );
		}

		else if( e->type == MMCP_DNY_SESSION_JOIN ) {
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
			err_ret( "MMCP, Session Join Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_start_join() : "
					 "Request for Session Join is Denied.", 0 );
		}

		else {
			/* Can't Enter Here. */
			mmcp_event_type_print( e->type );
			err_ret( "mmcp_ops_start_join() : Event type error.", 0 );
		}
	}	/* while( 1 ) */

	/* If Receives SJC with f = 1 ... */

	sjc = (mmcp_pkt_sjc *) e->packet;

	if( ep->role == MMCP_ROLE_LMC ) {

		/* Domain is IPv4. */
		if( mso->domain == AF_INET ) {
			memset( &lmc->group_addr, 0, sizeof(lmc->group_addr) );
			lmc->group_addr.sa.sin_family = AF_INET;
			lmc->group_addr.sa.sin_addr.s_addr = sjc->group_ip.ip;
			lmc->group_addr.sa.sin_port = sjc->group_port;
		}

		/* Domain is IPv6. */
		else if( mso->domain == AF_INET6 ) {
			memset( &lmc->group_addr, 0, sizeof(lmc->group_addr) );
			lmc->group_addr.sa6.sin6_family = AF_INET;
			memcpy( &lmc->group_addr.sa6.sin6_addr.s6_addr,
					&sjc->group_ip.ip6, sizeof(struct in6_addr) );
			lmc->group_addr.sa6.sin6_port = sjc->group_port;
		}

		/* Invoke Event Handler. */
		mso->hset.event_handler_is_running = 1;
		if( pthread_create(&mso->hset.event, NULL, 
							mso->hset.event_handler, mso) ) {
			err_ret( "mmcp_ops_start_join() : "
					 "cannot invoke event handler.", -1 );
		}

		/* SJR terminate. (LMC) */
		if( e ) {
			mmcp_destroy_event( e );
			e = NULL;
		}

		puts( "MMCP, Local Mobility Controller Join Completed! "
				"(STATE: ESTABLISHED)" );
		ep->state = MMCP_STATE_ESTABLISHED;

		/* Create SRT of LMC */
		mso->ep->lmc->sgt = mmcp_create_timer(
				MMCP_TIMER_SRT, MMCP_TIMER_SRT_VAL, NULL );
		mmcp_timerlist_add( &mso->timer_list, mso->ep->lmc->sgt );
		printf( "MMCP, LMC Creates SRT\n" );

		/* Create SPT of LMC */
		mso->ep->lmc->spt = mmcp_create_timer(
				MMCP_TIMER_SPT, MMCP_TIMER_SPT_VAL, NULL );
		mmcp_timerlist_add( &mso->timer_list, mso->ep->lmc->spt );
		printf( "MMCP, LMC Creates SPT\n" );
	}

	if( ep->role == MMCP_ROLE_MN ) {
		/* Domain is IPv4. */
		if( mso->domain == AF_INET ) {
			memset( &mn->lmc_addr, 0, sizeof(mn->lmc_addr) );
			mn->lmc_addr.sa.sin_family = AF_INET;
			mn->lmc_addr.sa.sin_addr.s_addr = sjc->lmc_ip.ip;
			mn->lmc_addr.sa.sin_port = sjc->lmc_port;

			memset( &mn->group_addr, 0, sizeof(mn->group_addr) );
			mn->group_addr.sa.sin_family = AF_INET;
			mn->group_addr.sa.sin_addr.s_addr = sjc->group_ip.ip;
			mn->group_addr.sa.sin_port = sjc->group_port;
		}

		/* Domain is IPv6. */
		else if( mso->domain == AF_INET6 ) {
			memset( &mn->lmc_addr, 0, sizeof(mn->lmc_addr) );
			mn->lmc_addr.sa6.sin6_family = AF_INET;
			memcpy( &mn->lmc_addr.sa6.sin6_addr.s6_addr,
					&sjc->lmc_ip.ip6, sizeof(struct in6_addr) );
			mn->lmc_addr.sa6.sin6_port = sjc->lmc_port;

			memset( &mn->group_addr, 0, sizeof(mn->group_addr) );
			mn->group_addr.sa6.sin6_family = AF_INET;
			memcpy( &mn->group_addr.sa6.sin6_addr.s6_addr,
					&sjc->group_ip.ip6, sizeof(struct in6_addr) );
			mn->group_addr.sa6.sin6_port = sjc->group_port;
		}

		/* SJR terminate. (MN) */
		if( e ) {
			mmcp_destroy_event( e );
			e = NULL;
		}

		if( !(ljr = mmcp_create_ljr(mso)) )
			err_ret( "mmcp_ops_start_join() : cannot create LJR.", -1 );

		addrlen = mmcp_addrlen(mso->domain);
		skb = mmcp_skb_alloc( mso, mso->msd, &mn->lmc_addr, addrlen );
		if( !skb )
			err_ret( "mmcp_ops_start_join() : mmcp_skb_alloc() failed.", -1 );

		skb->pktlen	= MMCP_PKT_LJR_LEN;
		skb->pkt	= ljr;

		mmcp_skblist_add( &mso->sndbuf, skb );

		/* Make Retransmission Timer (RXT) */
		rxt = mmcp_create_timer(MMCP_TIMER_RXT, MMCP_TIMER_RXT_VAL, skb );
		mmcp_timerlist_add( &mso->timer_list, rxt );

		mmcp_skb_transmit(skb);

		puts( "MMCP, Waiting Local Join Confirm! (STATE: LJR-SENT)" );
		ep->state = MMCP_STATE_LJR_SENT;

		while( 1 ) {

			e = mmcp_event_deq( &mso->joinq );

			/* Block: Waiting for LJC! */

			if( e->type == MMCP_CON_LOCAL_JOIN ) {
				mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
				puts( "MMCP, Local Join Confirm!" );
				break;
			}

			else if( e->type == MMCP_EXP_LOCAL_JOIN ) {
				/* RXT for SJR is already removed by mmcp_timer_dispatch(). */
				mmcp_destroy_event( e );
				ep->state = MMCP_STATE_CLOSED;
				err_ret( "MMCP, Local Join Failed! (STATE: CLOSED)\n"
						 "mmcp_ops_start_join() : "
						 "Exceed Maximum Retransmission for LJR.", -1 );
			}

			else if( e->type == MMCP_DNY_LOCAL_JOIN ) {
				mmcp_destroy_event( e );
				ep->state = MMCP_STATE_CLOSED;
				mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
				err_ret( "MMCP, Local Join Failed! (STATE: CLOSED)\n"
						 "mmcp_ops_start_join() : "
						 "Request for Local Join is Denied.", 0 );
			}

			else {
				/* Cannot Enter Here. */
				mmcp_event_type_print( e->type );
				err_ret( "mmcp_ops_start_join() : Event type error.", 0 );
			}
		}	/* while ( 1 ) */

		/* If Receives LJC with f = 1 ... */

		ljc = (mmcp_pkt_ljc *) e->packet;

		/* TODO */

		/* LJC Terminate. */
		if( e ) {
			mmcp_destroy_event( e );
			e = NULL;
		}

		/* Invoke Event Handler. */
		mso->hset.event_handler_is_running = 1;
		if( pthread_create(&mso->hset.event, NULL, 
							mso->hset.event_handler, mso) ) {
			err_ret( "mmcp_ops_start_join() : "
					 "cannot invoke event handler.", -1 );
		}

		/* Invoke Capture Handler. */
		mso->hset.capture_handler_is_running = 1;
		if( pthread_create(&mso->hset.capture, NULL, 
							mso->hset.capture_handler, mso) ) {
			err_ret( "mmcp_ops_start_join() : "
					 "cannot invoke event handler.", -1 );
		}

		puts( "MMCP, Local Join Completed! (STATE: ESTABLISHED)" );
		ep->state = MMCP_STATE_ESTABLISHED;

		/* Create SRT of MN */
		mso->ep->mn->sgt = mmcp_create_timer(
				MMCP_TIMER_SRT, MMCP_TIMER_SRT_VAL, NULL );
		mmcp_timerlist_add( &mso->timer_list, mso->ep->mn->sgt );
		printf( "MMCP, MN Create SRT\n" );
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////

int mmcp_ops_send_status_report( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_mn *mn = NULL;

	mmcp_pkt_usr *usr = NULL;

	if( ep->role != MMCP_ROLE_MN ) {
		err_ret( "mmcp_ops_status_monitoring(): Only MN.", -1 );
	}

	mn = mso->ep->mn;

	/* Make User Status Report Message */
	if( !(usr = mmcp_create_usr(mso)) )
		err_ret( "mmcp_ops_status_monitoring() : cannot create USR.", -1 );

//	printf( "usr->rcv_byte %d\n", usr->rcv_byte );
//	printf( "usr->elapsed_time %d\n", usr->elapsed_time );

	mmcp_pkt_transmit( mso->msd, usr, MMCP_PKT_USR_LEN,
			&mn->lmc_addr, mmcp_addrlen(mso->domain) );

//	puts( "MMCP, Send User Status Report!" );

	mn->sgt = mmcp_create_timer( MMCP_TIMER_SRT, MMCP_TIMER_SRT_VAL, NULL );
	mmcp_timerlist_add( &mso->timer_list, mn->sgt );
//	printf( "MMCP, MN Re-Creates SRT\n" );

	return 0;
}

int mmcp_ops_send_astatus_report( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_lmc *lmc = NULL;

	/* ASR Packet declaration */
	mmcp_pkt_asr *asr = NULL;

	if( ep->role != MMCP_ROLE_LMC ) {
		err_ret( "mmcp_ops_status_monitoring() : Only LMC.", -1 );
	}

	lmc = mso->ep->lmc;

	/* Make Aggregation Status Report Message */
	if( !(asr = mmcp_create_asr(mso)) )
		err_ret( "mmcp_ops_status_monitoring() : cannot create ASR.", -1 );

	mmcp_pkt_transmit( mso->msd, asr, MMCP_PKT_ASR_LEN,
						&lmc->sm_addr, mmcp_addrlen(mso->domain) );

	lmc->sgt = mmcp_create_timer( MMCP_TIMER_SRT, MMCP_TIMER_SRT_VAL, NULL );
	mmcp_timerlist_add( &mso->timer_list, lmc->sgt );

	return 0;
}

int mmcp_ops_send_status_probe( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;

	/* SM, LMC, MN declaration */
	struct mmcp_sm *sm = NULL;
	struct mmcp_lmc *lmc = NULL;

	/* USP Packet declaration */
	mmcp_pkt_usp *usp = NULL;

	int i;

	if( ep->role == MMCP_ROLE_LMC ) {
		lmc = mso->ep->lmc;

		for( i=0; i<4; i++ ) {
			if( lmc->spt_mn[i].flag == -1 )
				continue;

			else if( lmc->spt_mn[i].flag == 0 ) {
				switch( lmc->spt_mn[i].count ) {
					case 0:
					case 1:
					case 2:
						/* Make User Status Report Message */
						if( !(usp = mmcp_create_usp(mso)) )
							err_ret( "mmcp_ops_status_monitoring(): "
									 "cannot create USP.", -1 );

						mmcp_pkt_transmit( mso->msd, usp, MMCP_PKT_USP_LEN,
								&lmc->mn_addr[i], mmcp_addrlen(mso->domain) );

						lmc->spt_mn[i].count++;
						printf( "MMCP, LMC send User Status Probe to MN(%s)!\n",
							 		lmc->spt_mn[i].mn_id );
						break;

					case 3:
						lmc->spt_mn[i].flag = -1;
						lmc->spt_mn[i].count = 0;
						printf( "MMCP, MN(%s) regards termination\n",
							 	lmc->spt_mn[i].mn_id );
						break;

					default:
						break;
				}
			}

			else {
				lmc->spt_mn[i].flag = 0;
			}
		}

		lmc->spt = mmcp_create_timer(MMCP_TIMER_SPT, MMCP_TIMER_SPT_VAL, NULL);
		mmcp_timerlist_add( &mso->timer_list, lmc->spt );
//		printf( "MMCP, LMC Recreate SPT\n" );
	}

	if( ep->role == MMCP_ROLE_SM ) {
		sm = mso->ep->sm;

		for( i=0; i<4; i++ ) {
			if( sm->spt_lmc[i].flag == -1 )
				continue;

			else if( sm->spt_lmc[i].flag == 0 ) {
				switch( sm->spt_lmc[i].count ) {
					case 0:
					case 1:
					case 2:
						/* Make User Status Report Message */
						if( !(usp = mmcp_create_usp(mso)) )
							err_ret( "mmcp_ops_status_monitoring(): "
									 "cannot create USP.", -1 );

						mmcp_pkt_transmit( mso->msd, usp, MMCP_PKT_USP_LEN,
											&sm->lmclist[i].lmc_addr, 
											mmcp_addrlen(mso->domain) );

						sm->spt_lmc[i].count++;
						printf( "MMCP, SM send User Status Probe to LMC(%s)!\n",
							 	sm->spt_lmc[i].poa_id );
						break;

					case 3:
						sm->spt_lmc[i].flag = -1;
						sm->spt_lmc[i].count = 0;
						printf( "MMCP, LMC(%s) regards termination\n",
							 	sm->spt_lmc[i].poa_id );
						break;

					default:
						break;
				}
			}

			else {
				sm->spt_lmc[i].flag = 0;
			}
		}

		sm->spt = mmcp_create_timer(MMCP_TIMER_SPT, MMCP_TIMER_SPT_VAL, NULL);
		mmcp_timerlist_add( &mso->timer_list, sm->spt );
//		printf( "MMCP, SM Recreate SPT\n" );
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////

int mmcp_ops_user_leave( struct mmcp_socket *mso )
{
	return 0;
}

int mmcp_ops_accept_session_join( struct mmcp_socket *mso, mmcp_role_t role, 
									mmcp_addr_t *from )
{
	mmcp_pkt_sjc *sjc = NULL;

	/* Make SJC packet with F = 1. */
	if( !(sjc = mmcp_create_sjc(mso, 1)) )
		err_ret( "mmcp_ops_accept_session_join() : "
				 "mmcp_create_sjc() failed.", -1 );

	mmcp_pkt_transmit( mso->msd, sjc, MMCP_PKT_SJC_LEN, 
						from, mmcp_addrlen(mso->domain) );
	
	return 0;
}

int mmcp_ops_deny_session_join( struct mmcp_socket *mso, mmcp_role_t role, 
								mmcp_addr_t *from )
{
	return 0;
}

int mmcp_ops_accept_local_join( struct mmcp_socket *mso, mmcp_addr_t *from )
{
	mmcp_pkt_ljc *ljc = NULL;

	/* Make LJC packet with F = 1. */
	if( !(ljc = mmcp_create_ljc(mso, 1)) )
		err_ret( "mmcp_ops_accept_local_join() : "
				 "mmcp_create_ljc() failed.", -1 );

	mmcp_pkt_transmit( mso->msd, ljc, MMCP_PKT_LJC_LEN, 
						from, mmcp_addrlen(mso->domain) );

	printf( "Send LJC\n" );
	
	return 0;
}

int mmcp_ops_deny_local_join( struct mmcp_socket *mso, mmcp_addr_t *from )
{
	return 0;
}

int mmcp_ops_request_handover( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_mn *mn = mso->ep->mn;

	mmcp_pkt_hir *hir = NULL;
	mmcp_pkt_hic *hic = NULL;

	struct mmcp_skbuff *skb		= NULL;
	struct mmcp_timeritem *hit	= NULL;
	struct mmcp_eventitem *e	= NULL;

	socklen_t addrlen;
	char *str;

	if( !(ep->role == MMCP_ROLE_MN) )
		err_ret( "mmcp_ops_request_handover() : Only MN.", -1 );

	/* Make Handover Initiation Request Message. */
	if( !(hir = mmcp_create_hir(mso)) )
		err_ret( "mmcp_ops_request_handover() : cannot create HIR.", -1 );

	/* Make Send Buffer Item. */
	addrlen = mmcp_addrlen(mso->domain);
	skb = mmcp_skb_alloc( mso, mso->msd, &mn->lmc_addr, addrlen );

	str = inet_ntoa(mn->lmc_addr.sa.sin_addr);
//	printf( "lmc_addr: %s, %d\n", str, htons(mn->lmc_addr.sa.sin_port) );
//	printf( "hir type: %d\n", hir->basehdr.pkt_type );


	if( !skb )
		err_ret( "mmcp_ops_request_handover() : mmcp_skb_alloc() failed.", -1 );

	skb->pktlen	= MMCP_PKT_HIR_LEN;
	skb->pkt	= hir;

	/* TODO: knucsid 080303 - error check (hir packet - group_ip add ) */
	mmcp_skblist_add( &mso->sndbuf, skb );

	/* Make Retransmission Timer (RXT) */
	hit = mmcp_create_timer( MMCP_TIMER_HIT, MMCP_TIMER_HIT_VAL, skb );
	mmcp_timerlist_add( &mso->timer_list, hit );

	/* TODO: Make Handover Wating Timer (HWT) */

	mmcp_skb_transmit(skb);
//	mmcp_pkt_transmit( mso->msd, hir, MMCP_PKT_HIR_LEN,
//			&mn->lmc_addr, mmcp_addrlen(mso->domain) );

	puts( "MMCP, Waiting Handover Initiation Confirm! (STATE: HR-SENT)" );
	ep->state = MMCP_STATE_HR_SENT;

	while( 1 ) {

		e = mmcp_event_deq( &mso->handoverq );
		
		/* Block: Waiting for HIC! */
		if( !e ) {
			err_ret( "mmcp_ops_request_handover() : "
					 "NULL event occurs.", -1 );
		}

		else if( e->type == MMCP_CON_HANDOVER_INIT ) {
			puts( "MMCP, Handover Initiation Confirm!" );
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, hit));
			break;
		}

		else if( e->type == MMCP_DNY_HANDOVER_INIT ) {
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, hit));
			err_ret( "MMCP, Handover Init Request Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_request_handover() : "
					 "Request for Handover Initiation is Denied.", 0 );
		}

		else if( e->type == MMCP_EXP_HANDOVER_INIT ) {
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			err_ret( "MMCP, Handover Init Request Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_request_handover() : "
					 "Exceed Maximum Retransmission for HIR.", -1 );
		}

		else {
			/* Can't Enter Here. */
			mmcp_event_type_print( e->type );
			err_ret( "mmcp_ops_request_handover() : Event type error.", 0 );
		}
	}	/* while( 1 ) */

	/* If Receives HIC with f = 1 ... */

	hic = (mmcp_pkt_hic *) e->packet;

	/* Domain is IPv4. */
	if( mso->domain == AF_INET ) {
//		memset( &mn->lmc_addr, 0, sizeof(mn->lmc_addr) );
//		mn->lmc_addr.sa.sin_family = AF_INET;
//		mn->lmc_addr.sa.sin_addr.s_addr = hic->newlmc_ip.ip;
//		mn->lmc_addr.sa.sin_port = htons(hic->newlmc_port);

//	str = inet_ntoa(mn->lmc_addr.sa.sin_addr);
//	printf( "lmc_addr: %s, %d\n", str, htons(mn->lmc_addr.sa.sin_port) );

		memset( &mn->newlmc_addr, 0, sizeof(mn->newlmc_addr) );
		mn->newlmc_addr.sa.sin_family = AF_INET;
		mn->newlmc_addr.sa.sin_addr.s_addr = hic->newlmc_ip.ip;
		mn->newlmc_addr.sa.sin_port = htons(hic->newlmc_port);
	}

	/* Domain is IPv6. */
	else if( mso->domain == AF_INET6 ) {
		memset( &mn->newlmc_addr, 0, sizeof(mn->newlmc_addr) );
		mn->newlmc_addr.sa6.sin6_family = AF_INET6;
		memcpy( &mn->newlmc_addr.sa6.sin6_addr.s6_addr,
				&hic->newlmc_ip.ip6, sizeof(struct in6_addr) );
		mn->newlmc_addr.sa6.sin6_port = htons(hic->newlmc_port);
	}

	/* HIR terminate. */
	if( e ) {
		mmcp_destroy_event( e );
		e = NULL;
	}

	puts( "MMCP, Handover Initiation Completed! (STATE: ESTABLISHED)" );
	ep->state = MMCP_STATE_ESTABLISHED;
		
	return 0;
}

int mmcp_ops_transfer_handover( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_lmc *lmc = ep->lmc;

	mmcp_pkt_hct *hct = NULL;
	mmcp_pkt_hta *hta = NULL;
	mmcp_pkt_hic *hic = NULL;

	struct mmcp_skbuff *skb		= NULL;
	struct mmcp_timeritem *rxt	= NULL;
	struct mmcp_eventitem *e	= NULL;

	socklen_t addrlen;
	int i = 0;

	if( !(ep->role == MMCP_ROLE_LMC) )
		err_ret( "mmcp_ops_transfer_handover() : Only LMC!", -1 );


	/* Make Handover Context Transfer Message. */
	if( !(hct = mmcp_create_hct(mso)) )
		err_ret( "mmcp_ops_transfer_handover() : cannot create HCT.", -1 );

	/* Make Send Buffer Item. */
	addrlen = mmcp_addrlen(mso->domain);

	lmc = mso->ep->lmc;

	/* Temporary - knucsid 080226 */
	memcpy( &lmc->hd_val.newlmc_addr, &lmc->lmc_list[1].lmc_addr, addrlen );
	skb = mmcp_skb_alloc( mso, mso->msd, 
							&lmc->lmc_list[1].lmc_addr, addrlen );

	if( !skb ) {
		err_ret( "mmcp_ops_transfer_handover() : "
				 "mmcp_skb_alloc() failed.", -1 );
	}

	skb->pktlen	= MMCP_PKT_HCT_LEN;
	skb->pkt	= hct;

	mmcp_skblist_add( &mso->sndbuf, skb );

	/* Make Retransmission Timer (RXT) */
	rxt = mmcp_create_timer( MMCP_TIMER_RXT, MMCP_TIMER_RXT_VAL, skb );
	mmcp_timerlist_add( &mso->timer_list, rxt );

	mmcp_skb_transmit(skb);

	puts( "MMCP, Waiting Handover Transfer Acknowledgment!"
		  " (STATE: ESTABLISHED)" );

	while( 1 ) {

		e = mmcp_event_deq( &mso->handoverq );
		
		/* Block: Waiting for HTA! */

		if( e->type == MMCP_ACK_HANDOVER_TRANSFER ) {
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt) );
			puts( "LMC receives Handover Transfer Acknowledgment Packet." );

			break;
		}

		else if( e->type == MMCP_EXP_HANDOVER_TRANSFER ) {
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;

			err_ret( "MMCP, Handover Transfer Acknowledgment Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_transfer_handover() : "
					 "Exceed Maximum Retransmission for HCT.", 0 );

		}

		else {
			mmcp_event_type_print( e->type );
			err_ret( "mmcp_ops_transfer_handover() : Event type error.", 0 );
		}
	}	/* while( 1 ) */

	hta = (mmcp_pkt_hta *) e->packet;

	/* HCT terminate. */
	if( e ) {
		mmcp_destroy_event( e );
		e = NULL;
	}

	puts( "MMCP, Handover Transfer Completed! (STATE: ESTABLISHED)" );
	ep->state = MMCP_STATE_ESTABLISHED;
	
	/* Make HIC packet with F = 1. */
	if( !(hic = mmcp_create_hic(mso, 1)) )
		err_ret( "mmcp_ops_transfer_handover() : "
				 "mmcp_create_hic() failed.", -1 );

	for( i=0; i<4; i++ ) {
		if( strcmp(lmc->spt_mn[i].mn_id, &lmc->hd_val.mn_id) == 0 ) {
			lmc->spt_mn[i].flag = -1;
			break;
		}
	}

	mmcp_pkt_transmit( mso->msd, hic, MMCP_PKT_HIC_LEN, 
						&lmc->hd_val.hd_addr, mmcp_addrlen(mso->domain) );

	printf( "Send HIC\n" );

	return 0;
}

int mmcp_ops_transfer_ack_handover( struct mmcp_socket *mso, mmcp_addr_t *from )
{
	mmcp_pkt_hta *hta = NULL;

	if( !(hta = mmcp_create_hta(mso)) )
		err_ret( "mmcp_ops_transfer_ack_handover() : "
				 "mmcp_create_hta() failed.", -1 );

	mmcp_pkt_transmit( mso->msd, hta, MMCP_PKT_HTA_LEN, 
						from, mmcp_addrlen(mso->domain) );

	printf( "Send HTA\n" );
	
	return 0;
}

int mmcp_ops_igmp_send( struct mmcp_socket *mso )
{
	struct mmcp_lmc *lmc = mso->ep->lmc;

	int sock;
	struct ip_mreq		mreq;
	struct ipv6_mreq	mreq6;

	char *str;

	switch (mso->domain) {

	case AF_INET: 

		sock = socket( PF_INET, SOCK_DGRAM, 0 );
		if( sock == -1 )
			err_ret( "mmcp_ops_igmp_send() : Don't create IGMP socket", -1 );

		memcpy( &mreq.imr_multiaddr, &lmc->hd_val.group_addr.sa.sin_addr, 
				sizeof(struct in_addr) );

		mreq.imr_interface.s_addr = htonl(INADDR_ANY);

		str = inet_ntoa(mreq.imr_multiaddr);
//		printf( "mreq.imr_multiaddr : %s\n", str );

		return (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
							&mreq, sizeof(mreq)) );

	case AF_INET6:

		sock = socket( PF_INET6, SOCK_DGRAM, 0 );
		if( sock == -1 )
			err_ret( "mmcp_ops_igmp_send() : Don't create IGMP socket", -1 );

		memcpy( &mreq6.ipv6mr_multiaddr, &lmc->hd_val.group_addr.sa6.sin6_addr,
				sizeof(struct in6_addr) );

		mreq6.ipv6mr_interface = 0;

		return ( setsockopt(sock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
							&mreq6, sizeof(mreq6)) );

	default:
		errno = EAFNOSUPPORT;
		return -1;
	}
}

int mmcp_ops_local_join( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_mn *mn = mso->ep->mn;

	mmcp_pkt_ljr *ljr = NULL;
	mmcp_pkt_ljc *ljc = NULL;

	struct mmcp_skbuff *skb		= NULL;
	struct mmcp_timeritem *rxt	= NULL;
	struct mmcp_eventitem *e	= NULL;

	socklen_t addrlen;

	mmcp_ops_igmp_send(mso);

	if( !(ljr = mmcp_create_ljr(mso)) )
		err_ret( "mmcp_ops_local_join() : cannot create LJR.", -1 );

	addrlen = mmcp_addrlen(mso->domain);
	skb = mmcp_skb_alloc( mso, mso->msd, &mn->newlmc_addr, addrlen );
//	skb = mmcp_skb_alloc( mso, mso->msd, &mn->lmc_addr, addrlen );
	if( !skb )
		err_ret( "mmcp_ops_local_join() : mmcp_skb_alloc() failed.", -1 );

	skb->pktlen	= MMCP_PKT_LJR_LEN;
	skb->pkt	= ljr;

	mmcp_skblist_add( &mso->sndbuf, skb );

	/* Make Retransmission Timer (RXT) */
	rxt = mmcp_create_timer(MMCP_TIMER_RXT, MMCP_TIMER_RXT_VAL, skb );
	mmcp_timerlist_add( &mso->timer_list, rxt );

	mmcp_skb_transmit(skb);

	puts( "MMCP, Waiting Local Join Confirm! (STATE: LJR-SENT)" );
	ep->state = MMCP_STATE_LJR_SENT;

	while( 1 ) {

		e = mmcp_event_deq( &mso->joinq );

		/* Block: Waiting for LJC! */

		if( e->type == MMCP_CON_LOCAL_JOIN ) {
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
			puts( "MMCP, Local Join of Handover Confirm!" );
			break;
		}

		else if( e->type == MMCP_EXP_LOCAL_JOIN ) {
			/* RXT for SJR is already removed by mmcp_timer_dispatch(). */
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			err_ret( "MMCP, Local Join Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_local_join() : "
					 "Exceed Maximum Retransmission for LJR.", -1 );
		}

		else if( e->type == MMCP_DNY_LOCAL_JOIN ) {
			mmcp_destroy_event( e );
			ep->state = MMCP_STATE_CLOSED;
			mmcp_destroy_timer( mmcp_timerlist_del(&mso->timer_list, rxt));
			err_ret( "MMCP, Local Join Failed! (STATE: CLOSED)\n"
					 "mmcp_ops_local_join() : "
					 "Request for Local Join is Denied.", 0 );
		}

		else {
			/* Cannot Enter Here. */
			mmcp_event_type_print( e->type );
			err_ret( "mmcp_ops_local_join() : Event type error.", 0 );
		}
	}	/* while ( 1 ) */

	/* If Receives LJC with f = 1 ... */

	ljc = (mmcp_pkt_ljc *) e->packet;

	/* TODO */
//	memset( &mn->lmc_addr, 0, sizeof(mn->lmc_addr) );
	memcpy( &mn->lmc_addr, &mn->newlmc_addr, mmcp_addrlen(mso->domain) );

	/* LJC Terminate. */
	if( e ) {
		mmcp_destroy_event( e );
		e = NULL;
	}

	puts( "MMCP, Local Join Completed! (STATE: ESTABLISHED)" );
	ep->state = MMCP_STATE_ESTABLISHED;

	return 0;
}

/* ----------------------------------------------------------------- *
 *	Implementations of Synchronization Control Functions
 * ----------------------------------------------------------------- */

/* types.h --------------------------------------------------------- */

int mmcp_sem_init( mmcp_sem_t *sem, __u32 val )
{
	if( !sem )
		err_ret( "mmcp_sem_init() : null exception on \'sem\'.", -1 );

	sem->n = val;

	if( sem_init(&sem->s, 0, val) < 0 )
		err_ret( "mmcp_sem_init() : sem_init() failed.", -1 );

	return 0;
}

int mmcp_sem_destroy( mmcp_sem_t *sem )
{
	if( sem_destroy(&sem->s) < 0 )
		err_ret( "mmcp_sem_destroy() : sem_destroy() failed.", -1 );

	if( sem )
		free(sem);
	sem = NULL;

	return 0;
}


/* ----------------------------------------------------------------- *
 *	Get Interface Information Functions
 * ----------------------------------------------------------------- */

#define IF_BUFSIZE	1024
mn_status get_inter_io( char *_name )
{
	FILE *fp;
	char buf[IF_BUFSIZE];
	char name[30];
	char *lp;
	mn_status idev_info;
	interface_io ifio;
	int linenum = -2;

	if( (fp = fopen("/proc/net/dev", "r")) == NULL ) {
		perror( "get_inter_io() : fopen error" );
	}

	while( fgets(buf, IF_BUFSIZE-1, fp) != NULL ) {
		if( linenum < 0 ) {
			linenum++;
			continue;
		}

		sscanf( buf, "%s", name );
		lp = strstr( name, ":" );
		memset( (void *)&ifio, 0x00, sizeof(ifio) );

		if( strlen( lp+1 ) == 0 ) {
			sscanf( buf, "%s %d", name, &ifio.rcv_io.byte );
		}
		else {
			ifio.rcv_io.byte = atoi(lp+1);
			sscanf( buf, "%s", name );
		}

		lp = strstr( name, ":" );
		memcpy( ifio.dev_name, name, lp-name );
		idev_info.rcv_byte			= ifio.rcv_io.byte;
		idev_info.elapsed_time	= 0;

//		printf( "ifio.rcv_io.byte = %d\n", ifio.rcv_io.byte );
//		printf( "idev_info.rcv_byte = %d\n", idev_info.rcv_byte );
		if( strcmp( _name, ifio.dev_name) == 0 ) {
			break;
		}
	}

	fclose(fp);
	return idev_info;
}


/* ----------------------------------------------------------------- *
 *	Input/Output Control Functions
 * ----------------------------------------------------------------- */

/* mmcp.h ---------------------------------------------------------- */

mmcp_pkt_sjr *mmcp_create_sjr( struct mmcp_socket *mso )
{
	mmcp_pkt_sjr *sjr = NULL;

	sjr = (mmcp_pkt_sjr *) mmcp_pkt_alloc( MMCP_PKT_SJR, 0, 0 );
	if( !sjr )
		err_ret( "mmcp_create_sjr(): mmcp_pkt_alloc() failed.", NULL );

	memset( sjr, 0, MMCP_PKT_SJR );

	sjr->basehdr.pkt_type		= MMCP_PKT_SJR;
	sjr->basehdr.payload_leng	= MMCP_PKT_SJR_LEN;
	sjr->basehdr.f				= 0;
	sjr->basehdr.session_id		= 0;

	if( mso->ep->role == MMCP_ROLE_LMC ) {
		sjr->basehdr.mn_leng	= 0x0;
		memset( sjr->poa_id, 0, MMCP_POAID_LEN );
		memcpy( sjr->poa_id, mso->ep->lmc->poa_id, MMCP_POAID_LEN );
	}

	else if( mso->ep->role == MMCP_ROLE_MN ) {
		sjr->basehdr.mn_leng	= MMCP_MNID_LEN / 4;	/* Temp */
		memset( sjr->mn_id, 0, MMCP_MNID_LEN );
		memset( sjr->poa_id, 0, MMCP_POAID_LEN );
		memcpy( sjr->mn_id, mso->ep->mn->mn_id, MMCP_MNID_LEN );
		memcpy( sjr->poa_id, mso->ep->mn->poa_id, MMCP_POAID_LEN );
	}

	else {
		err_ret( "mmcp_create_sjr(): role does not exist.", NULL );
	}


	return sjr;
}

mmcp_pkt_sjc *mmcp_create_sjc( struct mmcp_socket *mso, int f )
{
	struct mmcp_sm *sm = mso->ep->sm;
	mmcp_pkt_sjc *sjc = NULL;

	sjc = (mmcp_pkt_sjc *) mmcp_pkt_alloc( MMCP_PKT_SJC, 0, 0 );
	if( !sjc )
		err_ret( "mmcp_create_sjc(): mmcp_pkt_alloc() failed.", NULL );

	memset( sjc, 0, MMCP_PKT_SJC );

	sjc->basehdr.pkt_type		= MMCP_PKT_SJC;
	sjc->basehdr.mn_leng		= 0x0;
	sjc->basehdr.payload_leng	= MMCP_PKT_SJC_LEN;
	sjc->basehdr.f				= f;
	sjc->basehdr.session_id		= 0;

	
	if( mso->domain == AF_INET ) {
		sjc->group_port		= sm->group_addr.sa.sin_port;
		sjc->group_ip.ip	= sm->group_addr.sa.sin_addr.s_addr;

		/* Temporary */
		sjc->lmc_port		= sm->lmclist[0].lmc_addr.sa.sin_port;
		sjc->lmc_ip.ip		= sm->lmclist[0].lmc_addr.sa.sin_addr.s_addr;
	}
	else if( mso->domain == AF_INET6 ) {
		sjc->group_port		= sm->group_addr.sa.sin_port;
		memcpy( &sjc->group_ip.ip6, &sm->group_addr.sa6.sin6_addr.s6_addr,
				sizeof(struct in6_addr) );

		/* Temporary */
		sjc->lmc_port		= sm->lmclist[0].lmc_addr.sa.sin_port;
		memcpy( &sjc->lmc_ip.ip6, 
				&sm->lmclist[0].lmc_addr.sa6.sin6_addr.s6_addr,
				sizeof(struct in6_addr) );
	}


	return sjc;
}

mmcp_pkt_ljr *mmcp_create_ljr( struct mmcp_socket *mso )
{
	mmcp_pkt_ljr *ljr = NULL;

	ljr = (mmcp_pkt_ljr *) mmcp_pkt_alloc( MMCP_PKT_LJR, 0, 0 );
	if( !ljr )
		err_ret( "mmcp_create_ljr(): mmcp_pkt_alloc() failed.", NULL );

	memset( ljr, 0, MMCP_PKT_LJR );

	ljr->basehdr.pkt_type		= MMCP_PKT_LJR;
	ljr->basehdr.payload_leng	= MMCP_PKT_LJR_LEN;
	ljr->basehdr.f				= 0;
	ljr->basehdr.session_id		= 0;
	ljr->basehdr.mn_leng		= MMCP_MNID_LEN / 4;
	memset( ljr->mn_id, 0, sizeof(ljr->mn_id) );
	memcpy( ljr->mn_id, mso->ep->mn->mn_id, MMCP_MNID_LEN );

	return ljr;
}

mmcp_pkt_ljc *mmcp_create_ljc( struct mmcp_socket *mso, int f )
{
	mmcp_pkt_ljc *ljc = NULL;

	ljc = (mmcp_pkt_ljc *) mmcp_pkt_alloc( MMCP_PKT_LJC, 0, 0 );
	if( !ljc )
		err_ret( "mmcp_create_ljc(): mmcp_pkt_alloc() failed.", NULL );

	memset( ljc, 0, MMCP_PKT_LJC );

	ljc->basehdr.pkt_type		= MMCP_PKT_LJC;
	ljc->basehdr.mn_leng		= 0x0;
	ljc->basehdr.payload_leng	= MMCP_PKT_LJC_LEN;
	ljc->basehdr.f				= f;
	ljc->basehdr.session_id		= 0;


	return ljc;
}

mmcp_pkt_ulr *mmcp_create_ulr( struct mmcp_socket *mso )
{
	mmcp_pkt_ulr *ulr = NULL;

	ulr = (mmcp_pkt_ulr *) mmcp_pkt_alloc( MMCP_PKT_ULR, 0, 0 );
	if( !ulr )
		err_ret( "mmcp_create_ulr(): mmcp_pkt_alloc() failed.", NULL );

	memset( ulr, 0, MMCP_PKT_ULR );

	/* TODO */
	ulr->basehdr.pkt_type = MMCP_PKT_ULR;


	return ulr;
}

mmcp_pkt_ulc *mmcp_create_ulc( struct mmcp_socket *mso )
{
	mmcp_pkt_ulc *ulc = NULL;

	ulc = (mmcp_pkt_ulc *) mmcp_pkt_alloc( MMCP_PKT_ULC, 0, 0 );
	if( !ulc )
		err_ret( "mmcp_create_ulc(): mmcp_pkt_alloc() failed.", NULL );

	memset( ulc, 0, MMCP_PKT_ULC );

	/* TODO */
	ulc->basehdr.pkt_type = MMCP_PKT_ULC;


	return ulc;
}

mmcp_pkt_usr *mmcp_create_usr( struct mmcp_socket *mso )
{
	struct mmcp_mn *mn = mso->ep->mn;
	mmcp_pkt_usr *usr = NULL;
	mn_status devinfo;
	struct timeval tmp_time;

	usr = (mmcp_pkt_usr *) mmcp_pkt_alloc( MMCP_PKT_USR, 0, 0 );
	if( !usr )
		err_ret( "mmcp_create_usr(): mmcp_pkt_alloc() failed.", NULL );

	memset( usr, 0, MMCP_PKT_USR );

	usr->basehdr.pkt_type		= MMCP_PKT_USR;
	usr->basehdr.payload_leng	= MMCP_PKT_USR_LEN;
	usr->basehdr.session_id		= 0;

	memcpy( usr->mn_id, mn->mn_id, MMCP_MNID_LEN );

	devinfo = get_inter_io( mn->ifname );

	usr->rcv_byte				= devinfo.rcv_byte - mn->init_byte;

	gettimeofday( &tmp_time, NULL );
	usr->elapsed_time			= tmp_time.tv_sec - mso->ep->mn->elapsed_time;

	return usr;
}

mmcp_pkt_asr *mmcp_create_asr( struct mmcp_socket *mso )
{
	mmcp_pkt_asr *asr = NULL;
	int i = 0;

	asr = (mmcp_pkt_asr *) mmcp_pkt_alloc( MMCP_PKT_ASR, 0, 0 );
	if( !asr )
		err_ret( "mmcp_create_asr(): mmcp_pkt_alloc() failed.", NULL );

	memset( asr, 0, MMCP_PKT_ASR );
	
	asr->basehdr.pkt_type		= MMCP_PKT_ASR;
	asr->basehdr.payload_leng	= MMCP_PKT_ASR_LEN;
	asr->basehdr.session_id		= 0;

	memcpy( asr->poa_id, mso->ep->lmc->poa_id, MMCP_POAID_LEN );
	for( i=0; i < 4; i++ ) {
		if( mso->ep->lmc->spt_mn[i].flag != -1 ) {
			memcpy( asr->mmcp_mn_info[i].mn_id, 
						mso->ep->lmc->mmcp_mn_info[i].mn_id, MMCP_MNID_LEN );
			asr->mmcp_mn_info[i].rcv_byte =
						mso->ep->lmc->mmcp_mn_info[i].rcv_byte;
			asr->mmcp_mn_info[i].elapsed_time =
						mso->ep->lmc->mmcp_mn_info[i].elapsed_time;

			/* Re-initiation */
			mso->ep->lmc->mmcp_mn_info[i].rcv_byte = 0;
			mso->ep->lmc->mmcp_mn_info[i].elapsed_time = 0;
		}
	}

	return asr;
}

mmcp_pkt_usp *mmcp_create_usp( struct mmcp_socket *mso )
{
	mmcp_pkt_usp *usp = NULL;

	usp = (mmcp_pkt_usp *) mmcp_pkt_alloc( MMCP_PKT_USP, 0, 0 );
	if( !usp )
		err_ret( "mmcp_create_usp(): mmcp_pkt_alloc() failed.", NULL );

	memset( usp, 0, MMCP_PKT_USP );

	usp->basehdr.pkt_type		= MMCP_PKT_USP;
	usp->basehdr.payload_leng	= MMCP_PKT_USP_LEN;
	usp->basehdr.session_id		= 0;


	return usp;
}

mmcp_pkt_hir *mmcp_create_hir( struct mmcp_socket *mso )
{
	mmcp_pkt_hir *hir = NULL;

	hir = (mmcp_pkt_hir *) mmcp_pkt_alloc( MMCP_PKT_HIR, 0, 0 );
	if( !hir )
		err_ret( "mmcp_create_hir(): mmcp_pkt_alloc() failed.", NULL );

	memset( hir, 0, MMCP_PKT_HIR );

	hir->basehdr.pkt_type		= MMCP_PKT_HIR;
	hir->basehdr.payload_leng	= MMCP_PKT_HIR_LEN;
	hir->basehdr.f				= 0;
	hir->basehdr.session_id		= 0;

/*	if( mso->domain == AF_INET )
	{
		hir->group_ip.ip = mso->ep->mn->group_addr.sa.sin_addr.s_addr;
	}

	else if( mso->domain == AF_INET6 )
	{
		memcpy( &hir->group_ip.ip6,
				&mso->ep->mn->group_addr.sa6.sin6_addr.s6_addr,
				sizeof(struct in6_addr) );
	}

	else
		err_ret( "mmcp_create_hir(): mso->domain error.", NULL );
*/
	hir->basehdr.mn_leng		= MMCP_MNID_LEN / 4;
	memset( hir->mn_id, 0, sizeof(hir->mn_id) );
	memcpy( hir->mn_id, mso->ep->mn->mn_id, MMCP_MNID_LEN );

	memset( hir->poa_id, 0, sizeof(hir->poa_id) );
	memcpy( hir->poa_id, mso->ep->mn->poa_id, MMCP_POAID_LEN );

	return hir;
}

mmcp_pkt_hct *mmcp_create_hct( struct mmcp_socket *mso )
{
	mmcp_pkt_hct *hct = NULL;

	hct = (mmcp_pkt_hct *) mmcp_pkt_alloc( MMCP_PKT_HCT, 0, 0 );
	if( !hct )
		err_ret( "mmcp_create_hct(): mmcp_pkt_alloc() failed.", NULL );

	memset( hct, 0, MMCP_PKT_HCT );

	hct->basehdr.pkt_type		= MMCP_PKT_HCT;
	hct->basehdr.payload_leng	= MMCP_PKT_HCT_LEN;
	hct->basehdr.f				= 0;
	hct->basehdr.session_id		= 0;

	hct->basehdr.mn_leng		= MMCP_MNID_LEN / 4;
	memset( hct->mn_id, 0, sizeof(hct->mn_id) );
	memcpy( hct->mn_id, mso->ep->lmc->hd_val.mn_id, MMCP_MNID_LEN );

	/* TODO: group ip setting - knucsid 080229 */
	if( mso->domain == AF_INET ) {
		memcpy( &hct->group_ip.ip,
				&mso->ep->lmc->group_addr.sa.sin_addr,
				sizeof(struct in_addr) );
	}

	if( mso->domain == AF_INET6 ) {
		memcpy( &hct->group_ip.ip6,
				&mso->ep->lmc->group_addr.sa6.sin6_addr, 
				sizeof(struct in6_addr) );
	}

//	char *str;
//	str = inet_ntoa( mso->ep->lmc->group_addr.sa.sin_addr);
//	printf( "lmc->group_addr %s\n", str );

	return hct;
}

mmcp_pkt_hta *mmcp_create_hta( struct mmcp_socket *mso )
{
	mmcp_pkt_hta *hta = NULL;

	hta = (mmcp_pkt_hta *) mmcp_pkt_alloc( MMCP_PKT_HTA, 0, 0 );
	if( !hta )
		err_ret( "mmcp_create_hta(): mmcp_pkt_alloc() failed.", NULL );

	memset( hta, 0, MMCP_PKT_HTA );

	hta->basehdr.pkt_type		= MMCP_PKT_HTA;
	hta->basehdr.mn_leng		= 0x0;
	hta->basehdr.payload_leng	= MMCP_PKT_HTA_LEN;
	hta->basehdr.f				= 0;
	hta->basehdr.session_id		= 0;

	return hta;
}

mmcp_pkt_hic *mmcp_create_hic( struct mmcp_socket *mso, int f )
{
	mmcp_pkt_hic *hic = NULL;

	hic = (mmcp_pkt_hic *) mmcp_pkt_alloc( MMCP_PKT_HIC, 0, 0 );
	if( !hic )
		err_ret( "mmcp_create_hic(): mmcp_pkt_alloc() failed.", NULL );

	memset( hic, 0, MMCP_PKT_HIC );

	hic->basehdr.pkt_type		= MMCP_PKT_HIC;
	hic->basehdr.mn_leng		= MMCP_MNID_LEN / 4;
	hic->basehdr.payload_leng	= MMCP_PKT_HIC_LEN;
	hic->basehdr.f				= f;
	hic->basehdr.session_id		= 0;

	memset( hic->mn_id, 0, sizeof(hic->mn_id) );
	memcpy( hic->mn_id, mso->ep->lmc->hd_val.mn_id, MMCP_MNID_LEN );
	hic->newlmc_port = htons(mso->ep->lmc->hd_val.newlmc_addr.sa.sin_port);
	// 4 of memcpy is temporary value - knucsid 080228
	memcpy( &hic->newlmc_ip, 
			&mso->ep->lmc->hd_val.newlmc_addr.sa.sin_addr.s_addr, 4 );

	return hic;
}

void *mmcp_pkt_alloc( mmcp_packet_t type, int length, int domain )
{
	void *pkt = NULL;
	size_t pktlen = 0;

	switch( type ) {

		case MMCP_PKT_SJR:
			pktlen = MMCP_PKT_SJR_LEN;
			break;

		case MMCP_PKT_SJC:
			pktlen = MMCP_PKT_SJC_LEN;
			break;

		case MMCP_PKT_LJR:
			pktlen = MMCP_PKT_LJR_LEN;
			break;

		case MMCP_PKT_LJC:
			pktlen = MMCP_PKT_LJC_LEN;
			break;

		case MMCP_PKT_ULR:
		case MMCP_PKT_ULC:
			break;

		case MMCP_PKT_USR:
			pktlen = MMCP_PKT_USR_LEN;
			break;

		case MMCP_PKT_ASR:
			pktlen = MMCP_PKT_ASR_LEN;
			break;

		case MMCP_PKT_USP:
			pktlen = MMCP_PKT_USP_LEN;
			break;

		case MMCP_PKT_HIR:
			pktlen = MMCP_PKT_HIR_LEN;
			break;

		case MMCP_PKT_HCT:
			pktlen = MMCP_PKT_HCT_LEN;
			break;

		case MMCP_PKT_HTA:
			pktlen = MMCP_PKT_HTA_LEN;
			break;

		case MMCP_PKT_HIC:
			pktlen = MMCP_PKT_HIC_LEN;
			break;

		default:
			break;
	}

	pkt = malloc(pktlen);
	if( !pkt )
		err_ret( "mmcp_pkt_alloc(): malloc() failed.", NULL );
	
	memset( pkt, 0, pktlen );

	return pkt;
}
