/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	event.c
 *	This source includes implementation of the MMCP event.
 *
 *	Contact	:	Jaesung Park (knucsid@gmail.com)
 *				Jiin Kim
 */

#include <header/mmcp.h>
#include <header/structs.h>
#include <header/event.h>
#include <header/wrapper.h>
#include <header/ieee802_11.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* ----------------------------------------------------------------- *
 *	Implementation fo Internal Event Loop & Handlers (Thread Functions).
 * ----------------------------------------------------------------- */

/* mmcp.h ---------------------------------------------------------- */

/*	mmcp_event_dispatch()
 *		SJR, SJC, LJR, LJC				: joinq
 *		ULR, ULC, USR, ASR, USP		 	: eventq	
 *		HIR, HCT, HTA, HIC				: handoverq
 */

void mmcp_event_dispatch( struct mmcp_socket *mso, fd_set *fds )
{
	struct mmcp_endpoint *ep = mso->ep;

	mmcp_addr_t from;
	socklen_t fromlen;
   
	if( mso->domain == AF_INET )
		fromlen	= SA_LEN;
	else if( mso->domain == AF_INET6 )
		fromlen = SS_LEN;

	/* Static buffer size : Base Header Length + Maximum Segment Size */
	__u32 buflen = MMCP_BASEHDR_LEN + 1500;	/* TODO */
	__u8 buf[buflen];
	int nbytes = -1;

	struct mmcp_basehdr *basehdr	= NULL;
	struct mmcp_eventitem *event	= NULL;

	if( FD_ISSET(mso->msd, fds) ) {

		nbytes = recvfrom( mso->msd, buf, buflen, 0, (SA *) &from, &fromlen );
		basehdr = (struct mmcp_basehdr *) buf;

//printf( "receive basehdr type  check : %d\n", basehdr->pkt_type );

		switch( basehdr->pkt_type ) {

		case MMCP_PKT_SJR:	/* SJR from MN or LMC to SM */

			puts( "MMCP, Receive SJR" );

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only SM) */
			if( ep->role != MMCP_ROLE_SM )
				err_goto( "mmcp_event_dispatch() : "
							"Only SM can receive SJR.", discard_event );

			/* If mn_leng == 0, LMC is requested. */
			if( basehdr->mn_leng == 0 ) {
				printf( "SJR packet received(LMC).\n" );

				event = mmcp_create_event( MMCP_REQ_SESSION_JOIN,
						MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );
			}

			/* If mn_leng != 0, MN is requested. */
			else if( basehdr->mn_leng != 0 ) {
				printf( "SJR packet received(MN).\n" );

				event = mmcp_create_event( MMCP_REQ_SESSION_JOIN,
						MMCP_ROLE_MN, &from, fromlen, buf, nbytes );
			}

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create SJR event.", 
							discard_event );

			if( mmcp_event_enq( &mso->joinq, event ) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_SJC:	/* SJC from SM to MN or LMC */

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_SJR_SENT )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (LMC or MN) */
			if( ep->role != MMCP_ROLE_LMC && ep->role != MMCP_ROLE_MN )
				err_goto( "mmcp_event_dispatch() : SM cannot receive SJC.",
						discard_event );

			/* SJR is denied by SM. (SJC with f = 0) */
			if( !basehdr->f ) {
				event = mmcp_create_event( MMCP_DNY_SESSION_JOIN, 
						MMCP_ROLE_SM, &from, fromlen, NULL, 0 );
			}

			/* SJR is accepted by SM. (SJC with f = 1) */
			else if( basehdr->f ) {
				event = mmcp_create_event( MMCP_CON_SESSION_JOIN,
						MMCP_ROLE_SM, &from, fromlen, buf, nbytes );
			}

			else {
				err_goto( "mmcp_event_dispatch() : f flag error.", 
							discard_event );
			}

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create SJC event.",
							discard_event );

			if( mmcp_event_enq(&mso->joinq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_LJR:	/* LJR from MN to LMC */

			printf( "LJR packet received(MN).\n" );

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only SM) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive LJR.",
							discard_event );

			event = mmcp_create_event( MMCP_REQ_LOCAL_JOIN,
					MMCP_ROLE_MN, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create LJR event.",
							discard_event );

			if( mmcp_event_enq(&mso->joinq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_LJC:	/* LJC from LJC to MN */

//			printf( "MN receives Local Join confirm.\n" );

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_LJR_SENT ) {
				printf( "ep->state = %d\n", ep->state );
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );
			}

			/* Check current role of endpoint. (Only MN) */
			if( ep->role != MMCP_ROLE_MN )
				err_goto( "mmcp_event_dispatch() : Only MN can receive LJC.",
							discard_event );

			/* LJR is denied by LMC. (LJC with f = 0) */
			if( !basehdr->f ) {
				event = mmcp_create_event( MMCP_DNY_LOCAL_JOIN,
						MMCP_ROLE_LMC, &from, fromlen, NULL, 0 );
			}

			/* LJR is accepted by LMC. (LJC with f = 1) */
			else if( basehdr->f ) {
				event = mmcp_create_event( MMCP_CON_LOCAL_JOIN,
						MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );
			}

			else {
				err_goto( "mmcp_event_dispatch() : f flag error.", 
							discard_event );
			}

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create LJC event.",
							discard_event );

			if( mmcp_event_enq(&mso->joinq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_ULR:	/* ULR from MN to LMC */

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only LMC) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive ULR.",
							discard_event );

			event = mmcp_create_event( MMCP_REQ_USER_LEAVE,
					MMCP_ROLE_MN, &from, fromlen, NULL, 0 );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create ULR event.",
							discard_event );

			if( mmcp_event_enq(&mso->eventq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_ULC:	/* ULC from LMC to MN */

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only MN) */
			if( ep->role != MMCP_ROLE_MN )
				err_goto( "mmcp_event_dispatch() : Only MN can receive ULC.",
							discard_event );

			/* ULR is denied by LMC. (ULC with f = 0) */
			if( !basehdr->f ) {
				event = mmcp_create_event( MMCP_DNY_USER_LEAVE,
						MMCP_ROLE_LMC, &from, fromlen, NULL, 0 );
			}

			/* ULR is accepted by LMC. (ULC with f = 1) */
			else if( basehdr->f ) {
				event = mmcp_create_event( MMCP_CON_USER_LEAVE,
						MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );
			}

			else {
				err_goto( "mmcp_event_dispatch() : f flag error.", 
							discard_event );
			}

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create ULC event.",
							discard_event );

			if( mmcp_event_enq(&mso->eventq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_USR:	/* USR from MN to LMC */

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only LMC) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive USR.",
							discard_event );


			event = mmcp_create_event( MMCP_RCV_STATUS_REPORT,
					MMCP_ROLE_MN, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create USR event.",
							discard_event );

			if( mmcp_event_enq(&mso->eventq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_ASR:	/* ASR from LMC to SM */
		
			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only SM) */
			if( ep->role != MMCP_ROLE_SM )
				err_goto( "mmcp_event_dispatch() : Only SM can receive ASR.",
							discard_event );

			event = mmcp_create_event( MMCP_RCV_ASTATUS_REPORT,
					MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create ASR sent.",
							discard_event );

			if( mmcp_event_enq(&mso->eventq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}


			break;

		case MMCP_PKT_USP:	/* USP from SM or LMC to LMC or MN */
			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED ) 
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint, (LMC or MN ) */
			if( ep->role != MMCP_ROLE_LMC && ep->role != MMCP_ROLE_MN )
				err_goto( "mmcp_event_dispatch() : LMC or MN can receive USP.",
							discard_event );
		
			else if( ep->role == MMCP_ROLE_MN ) {
				printf( "MMCP, MN Receives USP\n" );
				mmcp_timerlist_del( &mso->timer_list, ep->mn->sgt );
				event = mmcp_create_event( MMCP_SND_STATUS_REPORT,
						MMCP_ROLE_MN, &from, fromlen, buf, nbytes );	// ??
			}

			else if( ep->role == MMCP_ROLE_LMC ) {
				printf( "MMCP, LMC Receives USP\n" );
				mmcp_timerlist_del( &mso->timer_list, ep->lmc->sgt );
				event = mmcp_create_event( MMCP_SND_ASTATUS_REPORT,
						MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );	// ??
			}

			if( !event )
				err_goto( "mmcp_event_dispatch(): cannot create USP sent.",
							discard_event );

			if( mmcp_event_enq(&mso->eventq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch(): mmcp_event_enq() failed.",
							discard_event );
			}

			break;

		case MMCP_PKT_HIR:	/* HIR from MN to old LMC */

			puts( "MMCP, Receive HIR Packet." );

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only LMC) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive HIR.",
							discard_event );

			event = mmcp_create_event( MMCP_REQ_HANDOVER_INIT,
					MMCP_ROLE_MN, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create HIR event.",
							discard_event );

			if( mmcp_event_enq(&mso->handoverq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;


		case MMCP_PKT_HCT:	/* HCT from old LMC to new LMC */

			puts( "MMCP, Receive HCT" );

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only LMC) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive HCT.",
							discard_event );

			event = mmcp_create_event( MMCP_EXE_HANDOVER_TRANSFER,
					MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create HIR event.",
							discard_event );

			if( mmcp_event_enq(&mso->handoverq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;


		case MMCP_PKT_HTA:	/* HTA from new LMC to old LMC */

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_ESTABLISHED )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only LMC) */
			if( ep->role != MMCP_ROLE_LMC )
				err_goto( "mmcp_event_dispatch() : Only LMC can receive HTA.",
							discard_event );

			event = mmcp_create_event( MMCP_ACK_HANDOVER_TRANSFER,
					MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create HTA event.",
							discard_event );

			if( mmcp_event_enq(&mso->handoverq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;


		case MMCP_PKT_HIC:	/* HIC from old LMC to MN */
			puts("receive HIC");

			/* Check current state of endpoint. */
			if( ep->state != MMCP_STATE_HR_SENT )
				err_goto( "mmcp_event_dispatch() : State error.",
							discard_event );

			/* Check current role of endpoint. (Only MN) */
			if( ep->role != MMCP_ROLE_MN )
				err_goto( "mmcp_event_dispatch() : Only MN can receive HIC.",
							discard_event );

			/* HIR is denied by LMC. (HIC with f = 0) */
			if( !basehdr->f ) {
				event = mmcp_create_event( MMCP_DNY_HANDOVER_INIT, 
						MMCP_ROLE_LMC, &from, fromlen, NULL, 0 );
			}

			/* HIR is accepted by LMC. (HIC with f = 1) */
			else if( basehdr->f ) {
				event = mmcp_create_event( MMCP_CON_HANDOVER_INIT,
						MMCP_ROLE_LMC, &from, fromlen, buf, nbytes );
			}

			else {
				err_goto( "mmcp_event_dispatch() : f flag error.", 
							discard_event );
			}

			if( !event )
				err_goto( "mmcp_event_dispatch() : cannot create HIR event.",
							discard_event );

			if( mmcp_event_enq(&mso->handoverq, event) < 0 ) {
				mmcp_destroy_event( event );
				err_goto( "mmcp_event_dispatch() : mmcp_event_enq() failed.",
							discard_event );
			}

			break;


		default:
			break;

		}

	}

	return;

discard_event:
	memset( buf, 0, buflen );
	if( event )
		mmcp_destroy_event( event );

}	/* mmcp_event_dispatch() */



/* ----------------------------------------------------------------- *
 *	Implementation fo Internal Event Loop & Handlers (Thread Functions).
 * ----------------------------------------------------------------- */

/* event.h --------------------------------------------------------- */

void mmcp_handleset_init( struct mmcp_handleset *hset, mmcp_role_t role )
{
	hset->main_loop_is_running					= 0;
	hset->event_handler_is_running				= 0;
	hset->join_handler_is_running				= 0;
	hset->handover_handler_is_running			= 0;

	hset->capture_handler_is_running			= 0;

	hset->event_handler = mmcp_event_handler;

	if( role == MMCP_ROLE_SM || role == MMCP_ROLE_LMC ) {
		hset->join_handler = mmcp_join_handler;
		hset->handover_handler = mmcp_handover_handler;
	}

	if( role == MMCP_ROLE_MN || role == MMCP_ROLE_LMC ) {
		hset->main_loop		= mmcp_main_loop;
	}

	if( role == MMCP_ROLE_MN ) {
		hset->capture_handler = mmcp_capture_handler;
	}
}

void *mmcp_main_loop( void *arg )
{
	struct mmcp_socket *mso = (struct mmcp_socket *) arg;
	struct mmcp_handleset *hset = &mso->hset;

	struct timeval *to;
	struct timeval default_to;

	int result;
	fd_set backup_fds, read_fds;
	FD_ZERO(&backup_fds);
	FD_SET( mso->msd, &backup_fds );

	while( hset->main_loop_is_running ) {
		read_fds = backup_fds;

		if( !(to = mmcp_timerlist_peek_timeout(&mso->timer_list)) ) {
			default_to.tv_sec	= MMCP_MAINLOOP_TIMEOUT_SEC;
			default_to.tv_usec	= MMCP_MAINLOOP_TIMEOUT_USEC;
			to = &default_to;
		}

		result = select( mso->msd + 1, &read_fds, NULL, NULL, to );

		switch( result ) {
			case -1:
				result = 0;
				err_msg( "mmcp_event_loop() : select() failed." );
				break;
			case 0:
				mmcp_timer_dispatch(mso);
				break;
			default:
				mmcp_event_dispatch( mso, &read_fds );
				break;
		}	/* switch */
	}	/* while */

	return NULL;
}

void *mmcp_capture_handler( void *arg )
{
	struct mmcp_socket *mso = (struct mmcp_socket *) arg;

	wlan_ieee80211_capture( mso, MMCP_CAPTURE_DEV );
	pthread_exit(NULL);
}

void *mmcp_event_handler( void *arg )
{
	struct mmcp_socket *mso = (struct mmcp_socket *) arg;
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_handleset *hset = &mso->hset;

	struct mmcp_basehdr *basehdr = NULL;
	struct mmcp_eventitem *e = NULL;

	mmcp_pkt_usr *usr = NULL;
	mmcp_pkt_asr *asr = NULL;
	int i = 0, j = 0;

	while( hset->event_handler_is_running ) {

		e = mmcp_event_deq(&mso->eventq);

		if( !e ) {
			err_msg( "mmcp_event_handler() : NULL event occurs. [SKIP]" );
			continue;
		}

		else if( e->type == MMCP_SND_STATUS_REPORT ) {
			mmcp_ops_send_status_report(mso);
		}

		else if( e->type == MMCP_SND_ASTATUS_REPORT ) {
			mmcp_ops_send_astatus_report(mso);
		}

		else if( e->type == MMCP_SND_STATUS_PROBE ) {
			mmcp_ops_send_status_probe(mso);
		}

		else if( e->type == MMCP_RCV_STATUS_REPORT ) {
			usr = (mmcp_pkt_usr *) e->packet;
			for( i=0; i<4; i++ ) {

				if( ep->lmc->spt_mn[i].flag == -1 )
					continue;

				if( memcmp( usr->mn_id, ep->lmc->spt_mn[i].mn_id,
						 	sizeof(usr->mn_id) ) == 0 ) {

					ep->lmc->spt_mn[i].flag = 1;

					memcpy( ep->lmc->mmcp_mn_info[i].mn_id,
							usr->mn_id, MMCP_MNID_LEN );

					ep->lmc->mmcp_mn_info[i].rcv_byte = 
												usr->rcv_byte;
					ep->lmc->mmcp_mn_info[i].elapsed_time = 
												usr->elapsed_time;
					printf( "\n==== Status Information of MN ID: %s ====\n",
								usr->mn_id );
					printf( "Total number of packets received (byte): %d\n", 
								usr->rcv_byte );
					printf( "Elapsed Time: %d\n\n", usr->elapsed_time );
				}
			}
		}

		else if( e->type == MMCP_RCV_ASTATUS_REPORT ) {
			asr = (mmcp_pkt_asr *) e->packet;
			for( i=0; i<4; i++ ) {
				
				if( ep->sm->spt_lmc[i].flag == -1 )
					continue;

				if( memcmp( asr->poa_id, ep->sm->spt_lmc[i].poa_id,
						 	sizeof(asr->poa_id) ) == 0 ) {
					ep->sm->spt_lmc[i].flag = 1;
					ep->sm->spt_lmc[i].count = 0;

					printf( "\n==== Aggregation Status Information of "
							"PoA ID: %s ====\n", asr->poa_id );

					for( j=0; j<4; j++ ) {
						if( asr->mmcp_mn_info[j].elapsed_time != 0 ) {

						memcpy( ep->sm->mmcp_mn_info[i][j].mn_id, 
							asr->mmcp_mn_info[j].mn_id, MMCP_MNID_LEN );
						ep->sm->mmcp_mn_info[i][j].rcv_byte = 
							asr->mmcp_mn_info[j].rcv_byte;
						ep->sm->mmcp_mn_info[i][j].elapsed_time =
							asr->mmcp_mn_info[j].elapsed_time;

						printf( "MN Identification: %s\n",
								ep->sm->mmcp_mn_info[i][j].mn_id );
						printf( "Total number of packets received: %d\n",
								ep->sm->mmcp_mn_info[i][j].rcv_byte );
						printf( "Elapsed Time: %d\n\n",
								ep->sm->mmcp_mn_info[i][j].elapsed_time );
						}
					}
				}
			}
		}

//		else if( e->type == MMCP_RCV_STATUS_PROBE ) {
//			/* TODO: Data */
//		}

		else if( e->type == MMCP_REQ_USER_LEAVE ) {
			/* TODO */

		}
		else {
			err_msg( "mmcp_event_handler() : Event type error." );
		}

		basehdr = NULL;
		mmcp_destroy_event(e);
	} /* while */

	return NULL;
}

void *mmcp_join_handler( void *arg )
{
	struct mmcp_socket *mso = (struct mmcp_socket *) arg;
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_handleset *hset = &mso->hset;
	struct mmcp_sm *sm = mso->ep->sm;
	struct mmcp_lmc *lmc = mso->ep->lmc;

	struct mmcp_eventitem *e;
	mmcp_pkt_sjr *sjr = NULL;
	mmcp_pkt_ljr *ljr = NULL;

	int i=0;

	while( hset->join_handler_is_running ) {

		e = mmcp_event_deq( &mso->joinq );

		if( !e ) {
			err_msg( "mmcp_join_handler() : NULL join event occurs. [SKIP]" );
			continue;
		}

		/* Session Manager only */
		else if( ep->role == MMCP_ROLE_SM && 
					e->type == MMCP_REQ_SESSION_JOIN ) {

			puts( "Session Join Confirm Message Send!" );

			sjr = (mmcp_pkt_sjr *) e->packet;

			if( e->role == MMCP_ROLE_LMC ) {
				for( i=0; i<4; i++) {
					if( sm->spt_lmc[i].flag == -1 ) {
						memcpy( &sm->lmclist[i].lmc_addr, 
								&e->from, mmcp_addrlen(mso->domain) );
						memcpy( &sm->lmclist[i].poa_id,
								&sjr->poa_id, MMCP_POAID_LEN );
						memcpy( &sm->spt_lmc[i].poa_id,
								&sjr->poa_id, MMCP_POAID_LEN );
						sm->spt_lmc[i].flag = 0;
						break;
					}
				}

			}

			else if( e->role == MMCP_ROLE_MN ) {
				printf( "sjr->mn_id : %s\n", sjr->mn_id );
				printf( "sjr->poa_id : %s\n", sjr->poa_id );
			}


			mmcp_ops_accept_session_join( mso, e->role, &e->from );
			
		}

		/* Local Mobility Controller only */
		else if( ep->role == MMCP_ROLE_LMC &&
					e->type == MMCP_REQ_LOCAL_JOIN ) {

			puts( "Local Join Confirm Message Send!" );

			/* TODO: registration of MN - knucsid 080227 */
			ljr = (mmcp_pkt_ljr *) e->packet;

			for( i=0; i<4; i++ ) {
				if( lmc->spt_mn[i].flag == -1 ) {
					memcpy( &lmc->spt_mn[i].mn_id,
							&ljr->mn_id, MMCP_MNID_LEN );
					memcpy( &lmc->mn_addr[i], &e->from, 
							mmcp_addrlen(mso->domain) );
					lmc->spt_mn[i].flag = 0;
					lmc->spt_mn[i].count = 0;
					break;
				}
			}

			mmcp_ops_accept_local_join( mso, &e->from );

		}

		else {
			printf( "ep->role = %d, e->type = %d\n", ep->role, e->type );
			err_msg( "mmcp_join_handler() : Event type error." );
		}

		mmcp_destroy_event( e );

	}	/* while() */

	return NULL;
}

void *mmcp_handover_handler( void *arg )
{
	struct mmcp_socket *mso = (struct mmcp_socket *) arg;
	struct mmcp_endpoint *ep = mso->ep;
	struct mmcp_handleset *hset = &mso->hset;
	struct mmcp_lmc *lmc = mso->ep->lmc;

	struct mmcp_eventitem *e;
	mmcp_pkt_hir *hir = NULL;
	mmcp_pkt_hct *hct = NULL;

	while( hset->handover_handler_is_running ) {

		e = mmcp_event_deq( &mso->handoverq );

		if( !e ) {
			err_msg( "mmcp_handover_handler() : NULL handover event occurs." 
					 " [SKIP]" );
			continue;
		}

		else if( e->type == MMCP_REQ_HANDOVER_INIT ) {

			puts( "LMC receives Handover Initiation Request Packet." );
			
			hir = (mmcp_pkt_hir *) e->packet;

			/* Initiation of handover temporary value. */
			memset( &lmc->hd_val, 0, sizeof(lmc->hd_val) );

			if( mso->domain == AF_INET ) {
				memcpy( &lmc->hd_val.hd_addr.sa, &e->from, 
						mmcp_addrlen(mso->domain) );
			}

			if( mso->domain == AF_INET6 ) {
				memcpy( &lmc->hd_val.hd_addr.sa6, &e->from, 
						mmcp_addrlen(mso->domain) );
			}

			memcpy( &lmc->hd_val.mn_id, &hir->mn_id, MMCP_MNID_LEN ); 
			memcpy( &lmc->hd_val.poa_id, &hir->poa_id, MMCP_POAID_LEN ); 
//			printf( "lmc->hd_val.mn_id: %s\n",  lmc->hd_val.mn_id );
//			printf( "lmc->hd_val.poa_id: %s\n",  lmc->hd_val.poa_id );

			mmcp_ops_transfer_handover( mso );
		}

		else if( e->type == MMCP_EXE_HANDOVER_TRANSFER ) {

			puts( "New LMC receives Handover Context Transfer Packet." );

			hct = (mmcp_pkt_hct *) e->packet;

			if( mso->domain == AF_INET ) {
				memcpy( &lmc->hd_val.group_addr.sa.sin_addr, &hct->group_ip,
						sizeof(struct in_addr) );
			}

			if( mso->domain == AF_INET6 ) {
				memcpy( &lmc->hd_val.group_addr.sa6.sin6_addr, &hct->group_ip,
						sizeof(struct in6_addr) );
			}

//			mmcp_ops_igmp_send( mso );
//			puts( "Send IGMP" );

			mmcp_ops_transfer_ack_handover( mso, &e->from );


		}

		else {
			printf( "ep->role = %d, e->type = %d\n", ep->role, e->type );
			err_msg( "mmcp_handover_handler() : Event type error." );
		}

		mmcp_destroy_event( e );

	}	/* while() */

	return NULL;
}


/* ----------------------------------------------------------------- *
 *	Implementations of Internal Event Queue Manipulation Functions.
 * ----------------------------------------------------------------- */

void mmcp_event_type_print( mmcp_event_t type )
{
	if( type == MMCP_REQ_SESSION_JOIN )
		puts( "MMCP_REQ_SESSION_JOIN" );
	else if( type == MMCP_CON_SESSION_JOIN )
		puts( "MMCP_CON_SESSION_JOIN" );
	else if( type == MMCP_DNY_SESSION_JOIN )
		puts( "MMCP_DNY_SESSION_JOIN" );
	else if( type == MMCP_EXP_SESSION_JOIN )
		puts( "MMCP_EXP_SESSION_JOIN" );

	else if( type == MMCP_REQ_LOCAL_JOIN )
		puts( "MMCP_REQ_LOCAL_JOIN" );
	else if( type == MMCP_CON_LOCAL_JOIN )
		puts( "MMCP_CON_LOCAL_JOIN" );
	else if( type == MMCP_DNY_LOCAL_JOIN )
		puts( "MMCP_DNY_LOCAL_JOIN" );
	else if( type == MMCP_EXP_LOCAL_JOIN )
		puts( "MMCP_EXP_LOCAL_JOIN" );

	else if( type == MMCP_REQ_USER_LEAVE )
		puts( "MMCP_REQ_USER_LEAVE" );
	else if( type == MMCP_CON_USER_LEAVE )
		puts( "MMCP_CON_USER_LEAVE" );
	else if( type == MMCP_DNY_USER_LEAVE )
		puts( "MMCP_DNY_USER_LEAVE" );
	else if( type == MMCP_EXP_USER_LEAVE )
		puts( "MMCP_EXP_USER_LEAVE" );

	else if( type == MMCP_SND_STATUS_REPORT )
		puts( "MMCP_SND_STATUS_REPORT" );
	else if( type == MMCP_RCV_STATUS_REPORT )
		puts( "MMCP_RCV_STATUS_REPORT" );

	else if( type == MMCP_SND_ASTATUS_REPORT )
		puts( "MMCP_SND_ASTATUS_REPORT" );
	else if( type == MMCP_RCV_ASTATUS_REPORT )
		puts( "MMCP_RCV_ASTATUS_REPORT" );

	else if( type == MMCP_SND_STATUS_PROBE )
		puts( "MMCP_SND_STATUS_PROBE" );
	else if( type == MMCP_RCV_STATUS_PROBE )
		puts( "MMCP_RCV_STATUS_PROBE" );
	else if( type == MMCP_EXP_STATUS_PROBE )
		puts( "MMCP_EXP_STATUS_PROBE" );

	else if( type == MMCP_REQ_HANDOVER_INIT )
		puts( "MMCP_REQ_HANDOVER_INIT" );
	else if( type == MMCP_CON_HANDOVER_INIT )
		puts( "MMCP_CON_HANDOVER_INIT" );
	else if( type == MMCP_DNY_HANDOVER_INIT )
		puts( "MMCP_DNY_HANDOVER_INIT" );
	else if( type == MMCP_EXP_HANDOVER_INIT )
		puts( "MMCP_EXP_HANDOVER_INIT" );

	else if( type == MMCP_EXE_HANDOVER_TRANSFER )
		puts( "MMCP_EXE_HANDOVER_TRANSFER" );
	else if( type == MMCP_ACK_HANDOVER_TRANSFER )
		puts( "MMCP_ACK_HANDOVER_TRANSFER" );
	else if( type == MMCP_EXP_HANDOVER_TRANSFER )
		puts( "MMCP_EXP_HANDOVER_TRANSFER" );
}

struct mmcp_eventitem *mmcp_create_event( mmcp_event_t type, mmcp_role_t role,
		mmcp_addr_t *from, size_t fromlen, void *packet, __u32 packetlen )
{
	struct mmcp_eventitem *_new = NULL;

	/* TODO: knucsid 080303 - memory corruption check */
	if( !(_new = mmcp_ei_alloc()) )
		err_ret( "mmcp_create_event() : mmcp_ei_alloc() failed.", NULL );

	_new->type = type;
	_new->role = role;
	memcpy( &_new->from, from, fromlen );

	if( !packet || !packetlen )
		return _new;

	_new->packetlen = packetlen;
	_new->packet = malloc(packetlen);
	memcpy( _new->packet, packet, packetlen );

	return _new;
}

void mmcp_destroy_event( struct mmcp_eventitem *e )
{
	if( e->packetlen )
		free(e->packet);

	if( e )
		free(e);

	e = NULL;
}

int mmcp_event_initq( struct mmcp_eventq *q )
{
	if( mmcp_mutex_init(&q->mutex) < 0 )
		err_ret( "mmcp_event_initq() : mmcp_mutex_init() failed.", -1 );

	if( mmcp_sem_init(&q->sem, 0) < 0 ) {
		mmcp_mutex_destroy(&q->mutex);	// check
		err_ret( "mmcp_event_initq() : mmcp_sem_init() failed.", -1 );
	}

	q->head = q->tail = NULL;
	q->num = 0;

	return 0;
}

int mmcp_event_enq( struct mmcp_eventq *q, struct mmcp_eventitem *_new )
{
	mmcp_mutex_on(&q->mutex);

	if( mmcp_event_emptyq(q) ) {
		q->head = q->tail = _new;
		_new->next = _new->prev = _new;
	}
	else {
		mmcp_ei_add( _new, q->tail, q->head );
		q->tail = _new;
	}

	q->num++;
	_new->q = q;

	mmcp_mutex_off(&q->mutex);
	mmcp_sem_post(&q->sem);

	return 0;
}

struct mmcp_eventitem *mmcp_event_deq( struct mmcp_eventq *q )
{
	struct mmcp_eventitem *entry = NULL;

	mmcp_sem_wait(&q->sem);		/* blocking */

	if( mmcp_event_emptyq(q) )
		return NULL;

	mmcp_mutex_on(&q->mutex);

	q->head = q->head->next;
	mmcp_ei_del( entry = q->head->prev );

	q->num--;

	if( mmcp_event_emptyq(q) )
		q->head = q->tail = NULL;

	mmcp_mutex_off(&q->mutex);

	return entry;
}
