/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	structs.h
 *	This file is the implementation of the MMCP internal functions for
 *	Sockets API.
 *
 *	Contact :	Jaesung Park (knucsid@gmail.com)
 *				Jiin Kim
 */

#include <header/mmcp.h>
#include <header/structs.h>
#include <header/types.h>
#include <header/wrapper.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>				/* for malloc(), calloc(), free() */
#include <unistd.h>

//#include <sys/ioctl.h>
//#include <netinet/ether.h>
//#include <net/if.h>


/* ----------------------------------------------------------------- *
 *	Definitions for MMCP Global Variables
 * ----------------------------------------------------------------- */

/* structs.h ------------------------------------------------------- */

/* MMCP Global Variable */
struct mmcp_globals		*mmcp_global = NULL;

/* Socket Pointer that is used to Current Process */
struct mmcp_socket		*mmcp_current = NULL;

struct mmcp_globals *mmcp_create_globals()
{
	struct mmcp_globals *_new = NULL;

	_new = (struct mmcp_globals *) malloc(sizeof(struct mmcp_globals));
	if( !_new )
		err_ret( "mmcp_create_globals() : malloc() failed.", NULL );

	if( mmcp_mutex_init(&_new->mutex) < 0 ) {
		free(_new);
		err_ret( "mmcp_create_globals() : mmcp_mutex_init() failed.", NULL );
	}

	_new->nsock = 0;	/* 070215 knucsid */

	return _new;
}

void mmcp_destroy_globals( struct mmcp_globals *global )
{
	/* TODO */
}



/* ----------------------------------------------------------------- *
 *	Internal function of MMCP Socket
 * ----------------------------------------------------------------- */

/* structs.h ------------------------------------------------------- */

struct mmcp_socket *mmcp_create_socket( int domain, int protocol,
												mmcp_role_t role )
{
	struct mmcp_socket *_new = NULL;

	/* Initialize Global Variables. */
	if( !mmcp_global )
		mmcp_global = mmcp_create_globals();

	/* Check Boundary Condition. */
	if( mmcp_global->nsock >= MMCP_MAX_SOCKETS ) {
		printf( "mmcp_global->nsock = %d\n", mmcp_global->nsock );
		err_ret( "mmcp_create_socket() : cannot create anymore.", NULL );
	}

	/* Allocate memory for a new socket. */
	if( !(_new = mmcp_sk_alloc()) )
		err_ret( "mmcp_create_socket() : mmcp_sk_alloc() failed.", NULL );

	/* Initialize the interal fields and open new socket. */
	if( mmcp_sk_init(_new, domain, protocol, role) < 0 ) {
		mmcp_sk_free(_new);
		err_ret( "mmcp_create_socket() : mmcp_sk_init() failed.", NULL );
	}

	/* Add Current Socket to a Global Array/List. */
	mmcp_mutex_on( &mmcp_global->mutex );

	mmcp_global->socks[mmcp_global->nsock++] = _new;

	if( mmcp_mutex_off(&mmcp_global->mutex) < 0 ) {
		mmcp_sk_free(_new);
		mmcp_global->nsock--;
		err_ret( "mmcp_create_socket() : mmcp_mutex_off() failed.", NULL );
	}

	/* Set The Current Pointer. */
	mmcp_current = _new;

	return _new;
}

int mmcp_destroy_socket( struct mmcp_socket *mso )
{
	close(mso->msd);
	mmcp_sk_free(mso);

	return 0;
}

struct mmcp_socket *mmcp_sk_alloc()
{
	struct mmcp_socket *mso = NULL;

	mso = (struct mmcp_socket *) malloc(MMCP_SOCKET_LEN);
	if( !mso )
		err_ret( "mmcp_sk_alloc() : malloc() failed.", NULL );

	return mso;
}

void mmcp_sk_free( struct mmcp_socket *mso )
{
	if( mso )
		free(mso);

	mso = NULL;
}

int mmcp_sk_init( struct mmcp_socket *mso, int domain, int protocol,
							mmcp_role_t role )
{
	if( protocol == IPPROTO_UDP ) {
		if( (mso->msd = socket(domain, SOCK_DGRAM, IPPROTO_UDP)) < 0 ) {
			err_goto( "mmcp_sk_init() : socket() failed.", close_socket ); 
			
		}
	}
	
	else if( protocol == IPPROTO_TCP ) {
		if( (mso->msd = socket(domain, SOCK_STREAM, IPPROTO_TCP)) < 0 ) {
			err_goto( "mmcp_sk_init() : socket() failed.", close_socket ); 
		}
	}

	else {
		err_ret( "mmcp_sk_init() : No match protocol.", -1 ); 
	}

	mmcp_global->nsock = mso->msd;
	mso->domain = domain;
	mso->protocol = protocol;
	mso->ep = NULL;

	/* Initialize timer list. */
	if( mmcp_timerlist_init(&mso->timer_list) < 0 )
		err_goto( "mmcp_sk_init() : "
					"cannot initialize timer list.", close_socket );

	/* Initialize event queue. */
	if( mmcp_event_initq(&mso->joinq) < 0 )
		err_goto( "mmcp_sk_init() : "
				"cannot initialize join queue.", close_socket );

	if( mmcp_event_initq(&mso->handoverq) < 0 )
		err_goto( "mmcp_sk_init() : "
				"cannot initialize handover queue.", close_socket );

	if( mmcp_event_initq(&mso->eventq) < 0 )
		err_goto( "mmcp_sk_init() : "
				"cannot initialize event queue.", close_socket );

	/* Initialize event handler. */
	mmcp_handleset_init( &mso->hset, role );

	mso->max_sndbuf = MMCP_PARAM_SNDBUF_DFL;

	/* Initialize the send buffer. */
	if( mmcp_skblist_init(&mso->sndbuf, MMCP_PARAM_SNDBUF_DFL) < 0 )
		err_goto( "mmcp_sk_init() : "
					"mmcp_skblist_init() failed.", close_socket );

	return 0;

close_socket:
	close(mso->msd);
	return -1;
}

int mmcp_sk_bind( struct mmcp_socket *mso, struct sockaddr_storage *addr )
{
	int addrlen;

	addrlen = mmcp_addrlen( mso->domain );
	memcpy( &mso->addr, addr, addrlen );

	if( mso->ep->role != MMCP_ROLE_MN ) {

		if( mso->domain == AF_INET ) {
			if( bind(mso->msd, (SA *) &mso->addr, addrlen) < 0 ) {
				err_ret( "mmcp_sk_bind() : IPv4 bind() failed.", -1 );
			}
		}
		else if( mso->domain == AF_INET6 ) {
			if( bind(mso->msd, (SA *) &mso->addr, addrlen) < 0 ) {
				err_ret( "mmcp_sk_bind() : IPv6 bind() failed.", -1 );
			}
		}
		else {
			err_ret( "mmcp_sk_bind() : No domain.", -1 );
		}

	}

	return 0;
}


/* ----------------------------------------------------------------- *
 *	Internal function of MMCP Endpoint
 * ----------------------------------------------------------------- */

/* structs.h ------------------------------------------------------- */

int mmcp_create_endpoint( struct mmcp_socket *mso, mmcp_role_t role )
{
	struct mmcp_endpoint *ep = NULL;

	/* Allocate memory for a new endpoint. */
	if( !(ep = mmcp_ep_alloc()) < 0 )
		err_ret( "mmcp_create_endpoint() : mmcp_ep_alloc() failed.", -1 );

	ep->mso = mso;
	ep->role = role;

	if( mmcp_ep_init(mso, ep) < 0 )
		err_ret( "mmcp_create_endpoint() : mmcp_ep_init() failed.", -1 );

	ep->state = MMCP_STATE_CLOSED;
	mso->ep = ep;

	return 0;
}

int mmcp_destroy_endpoint( struct mmcp_endpoint *ep )
{
	mmcp_ep_free(ep);

	return 0;
}

struct mmcp_endpoint *mmcp_ep_alloc()
{
	struct mmcp_endpoint *ep = NULL;

	ep = (struct mmcp_endpoint *) malloc(MMCP_ENDPOINT_LEN);
	if( !ep )
		err_ret( "mmcp_ep_alloc() : malloc() failed.", NULL );

	return ep;
}

void mmcp_ep_free( struct mmcp_endpoint *ep )
{
	if( ep->role == MMCP_ROLE_SM )
		mmcp_sm_free(ep->sm);

	if( ep->role == MMCP_ROLE_LMC )
		mmcp_lmc_free(ep->lmc);

	if( ep->role == MMCP_ROLE_MN )
		mmcp_mn_free(ep->mn);

	if( ep )
		free(ep);

	ep = NULL;
}

int mmcp_ep_init( struct mmcp_socket *mso, struct mmcp_endpoint *ep )
{
	if( ep->role == MMCP_ROLE_SM ) {
		if( !(ep->sm = mmcp_sm_alloc()) < 0 )
			err_ret( "mmcp_ep_init() : mmcp_sm_alloc() failed.", -1 );

		mmcp_sm_init( mso, ep->sm );
	}

	else if( ep->role == MMCP_ROLE_LMC ) {
		if( !(ep->lmc = mmcp_lmc_alloc()) < 0 )
			err_ret( "mmcp_ep_init() : mmcp_lmc_alloc() failed.", -1 );

		mmcp_lmc_init( ep->lmc );
	}

	else if( ep->role == MMCP_ROLE_MN ) {
		if( !(ep->mn = mmcp_mn_alloc()) < 0 )
			err_ret( "mmcp_ep_init() : mmcp_mn_alloc() failed.", -1 );

		mmcp_mn_init( ep->mn );
	}
	else
		err_ret( "mmcp_ep_init() : the role does not exist.", -1 );

	return 0;
}


/* ----------------------------------------------------------------- *
 *	Internal function of Session Manager (SM)
 * ----------------------------------------------------------------- */

struct mmcp_sm *mmcp_sm_alloc()
{
	struct mmcp_sm *sm = NULL;

	sm = (struct mmcp_sm *) malloc(MMCP_SM_LEN);
	if( !sm )
		err_ret( "mmcp_sm_alloc() : malloc() failed.", NULL );

	return sm;
}

void mmcp_sm_free( struct mmcp_sm *sm )
{
	if( sm )
		free(sm);

	sm = NULL;
}

void mmcp_sm_init( struct mmcp_socket *mso, struct mmcp_sm *sm )
{
	int i = 0;

	sm->mn_index = 0;
	for( i=0; i<4; i++ ) {
		sm->spt_lmc[i].count = 0;
		sm->spt_lmc[i].flag = -1;
		memset( sm->spt_lmc[i].poa_id, 0, sizeof(sm->spt_lmc[i].poa_id ) );
	}

	sm->spt = mmcp_create_timer( MMCP_TIMER_SPT, MMCP_TIMER_SPT_VAL, NULL );
	mmcp_timerlist_add( &mso->timer_list, sm->spt );
	puts( "MMCP, SM creates SPT" );
}


/* ----------------------------------------------------------------- *
 *	Internal function of Local Mobility Controller (LMC)
 * ----------------------------------------------------------------- */

struct mmcp_lmc *mmcp_lmc_alloc()
{
	struct mmcp_lmc *lmc = NULL;

	lmc = (struct mmcp_lmc *) malloc(MMCP_LMC_LEN);
	if( !lmc )
		err_ret( "mmcp_lmc_alloc() : malloc() failed.", NULL );

	return lmc;
}

void mmcp_lmc_free( struct mmcp_lmc *lmc )
{
	if( lmc )
		free(lmc);

	lmc = NULL;
}

void mmcp_lmc_init( struct mmcp_lmc *lmc )
{
	int i = 0;

	lmc->lmc_list[0].lmc_addr.sa.sin_family = AF_INET;
	lmc->lmc_list[0].lmc_addr.sa.sin_addr.s_addr = inet_addr("192.168.1.2");
	lmc->lmc_list[0].lmc_addr.sa.sin_port = htons(MMCP_LMC_PORT_ONE);
	memcpy( lmc->lmc_list[0].poa_id, MMCP_POA_ONE, sizeof(MMCP_POA_ONE) );
	lmc->lmc_list[1].lmc_addr.sa.sin_family = AF_INET;
	lmc->lmc_list[1].lmc_addr.sa.sin_addr.s_addr = inet_addr("192.168.2.2");
	lmc->lmc_list[1].lmc_addr.sa.sin_port = htons(MMCP_LMC_PORT_TWO);
	memcpy( lmc->lmc_list[1].poa_id, MMCP_POA_TWO, sizeof(MMCP_POA_TWO) );

	for( i=0; i<4; i++ ) {
		lmc->spt_mn[i].count = 0;
		lmc->spt_mn[i].flag = -1;
		memset( lmc->spt_mn[i].mn_id, 0, sizeof(lmc->spt_mn[i].mn_id ) );
	}
}


/* ----------------------------------------------------------------- *
 *	Internal function of Mobile Node (MN)
 * ----------------------------------------------------------------- */

struct mmcp_mn *mmcp_mn_alloc()
{
	struct mmcp_mn *mn = NULL;

	mn = (struct mmcp_mn *) malloc(MMCP_MN_LEN);
	if( !mn )
		err_ret( "mmcp_mn_alloc() : malloc() failed.", NULL );

	return mn;
}

void mmcp_mn_free( struct mmcp_mn *mn )
{
	if( mn )
		free(mn);

	mn = NULL;
}

void mmcp_mn_init( struct mmcp_mn *mn )
{
	mn_status dev;

	system( "./init.sh" );

	mn->ifname = malloc( sizeof(MMCP_USE_DEV) );
	strncpy( mn->ifname, MMCP_USE_DEV, sizeof(MMCP_USE_DEV) );

	dev = get_inter_io( mn->ifname );
	mn->init_byte	= dev.rcv_byte;
//	printf( "mn->init_byte: %d\n", mn->init_byte );
//	printf( "mn->init_packet: %d\n", mn->init_packet );

	mn->elapsed_time = 0;
		
	memset( mn->mn_id, 0, sizeof(mn->mn_id) );
	memcpy( mn->mn_id, MMCP_MN_ID, sizeof(MMCP_MN_ID) );
	memset( mn->poa_id, 0, sizeof(mn->poa_id) );
	memcpy( mn->poa_id, MMCP_POA_ONE, sizeof(MMCP_POA_ONE) );
}

/* ----------------------------------------------------------------- *
 *	Implementation of Functinos for the MMCP Socket Buffer List
 * ----------------------------------------------------------------- */

int mmcp_skblist_init( struct mmcp_skblist *list, __u32 max_length )
{
	if( mmcp_mutex_init(&list->mutex) < 0 ) {
		err_ret( "mmcp_skblist_init() : mmcp_mutex_init() failed.", -1 );
	}

	if( mmcp_sem_init(&list->sem, 0) < 0 ) {
		err_ret( "mmcp_skblist_init() : mmcp_sem_init() failed.", -1 );
		mmcp_mutex_destroy(&list->mutex);
	}

	list->head = list->tail = NULL;
	list->num = list->length = 0;
	list->max_length = max_length;

	return 0;
}

int mmcp_skblist_add( struct mmcp_skblist *list, struct mmcp_skbuff *_new )
{
	mmcp_mutex_on(&list->mutex);

	if( mmcp_skblist_empty(list) ) {
		list->head = list->tail = _new;
		_new->next = _new->prev = _new;
	}
	else {
		if( mmcp_skblist_spare(list) < _new->pktlen ) {
			mmcp_mutex_off(&list->mutex);
			err_ret( "mmcp_skblist_add() : cannot add anymore.", -1 );
		}

		/* add_tail() */
		mmcp_skb_add( _new, list->tail, list->head );
		list->tail = _new;
	}

	list->num++;
	list->length += _new->pktlen;
	_new->list = list;

	mmcp_mutex_off(&list->mutex);

	mmcp_sem_post(&list->sem);

	return 0;
}

struct mmcp_skbuff *mmcp_skblist_del( struct mmcp_skblist *list,
									  struct mmcp_skbuff *skb )
{
	struct mmcp_skbuff *entry = NULL;

	mmcp_sem_wait(&list->sem);

	mmcp_mutex_on(&list->mutex);

	if( !skb ) {
		list->head = list->head->next;
		mmcp_skb_unlink( entry = list->head->prev );
		list->num--;
		list->length -= entry->pktlen;
	}
	else {
		if( skb == list->head )
			list->head = skb->next;
		else if( skb == list->tail )
			list->tail = skb->prev;

		mmcp_skb_unlink( entry = skb );
		list->num--;
		list->length -= entry->pktlen;
	}

	if( mmcp_skblist_empty(list) )
		list->head = list->tail = NULL;

	if( mmcp_mutex_off(&list->mutex) < 0 )
		err_ret( "mmcp_skblist_del() : mmcp_mutex_off() failed.", entry );

	return entry;
}
