/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	timer.c
 *	This source includes implementation of the MMCP timer.
 *
 *	Contact	:	Jaesung Park (knucsid@gmail.com)
 *				Jiin Kim
 */

#include <header/mmcp.h>
#include <header/structs.h>
#include <header/event.h>
#include <header/wrapper.h>

#include <sys/time.h>	/* gettimeofday() */
#include <time.h>

/* ----------------------------------------------------------------- *
 *	Implementation of MMCP Timer Event Manipulation Functions.
 * ----------------------------------------------------------------- */

void mmcp_ti_free( struct mmcp_timeritem *alarm )
{
	if( alarm->skb ) {
		mmcp_skblist_del( alarm->skb->list, alarm->skb );
		mmcp_skb_free(alarm->skb);
	}

	free(alarm);
	alarm = NULL;
}

int mmcp_timerlist_init( struct mmcp_timerlist *list )
{
	if( mmcp_mutex_init(&list->mutex) < 0 )
		err_ret( "mmcp_timerlist_init() : mmcp_mutext_init() failed.", -1 );

	list->head = NULL;
	list->tail = NULL;
	list->num = 0;

	return 0;
}

int mmcp_timerlist_add( struct mmcp_timerlist *list, 
						struct mmcp_timeritem *_new )
{
	struct mmcp_timeritem *t;
	__u32 num = list->num;

	mmcp_mutex_on(&list->mutex);

	if( mmcp_timerlist_empty(list) ) {
		list->head = list->tail = _new;
		_new->next = _new->prev = _new;
	}
	else {
		t = list->tail;
		while( num > 0 ) {
			if( (t->action_time.tv_sec < _new->action_time.tv_sec) ||
					((t->action_time.tv_sec == _new->action_time.tv_sec) &&
					 (t->action_time.tv_usec < _new->action_time.tv_usec)) ) {
				break;
			}

			t = t->prev;
			num--;
		}

		if( num == list->num ) {
			mmcp_ti_add( _new, list->tail, list->head );
			list->tail = _new;
		}
		else if( !num ) {
			mmcp_ti_add( _new, list->tail, list->head );
			list->head = _new;
		}
		else
			mmcp_ti_add( _new, t, t->next );
	}

	list->num++;
	_new->list = list;

	if( mmcp_mutex_off(&list->mutex) < 0 ) {
		mmcp_ti_unlink(_new);
		err_ret( "mmcp_timerlist_add() : mmcp_mutex_off() failed.\n"
				 "                     : Check failed timer in the list.", -1 );
	}

	return 0;
}

struct mmcp_timeritem *mmcp_timerlist_del( struct mmcp_timerlist *list,
										   struct mmcp_timeritem *alarm )
{
	struct mmcp_timeritem *entry = NULL;

	if( mmcp_timerlist_empty(list) )
		err_ret( "mmcp_timerlist_del() : timer list is empty.", NULL );

	mmcp_mutex_on(&list->mutex);

	/* blocking ... */


	if( !alarm ) {
		list->head = list->head->next;
		entry = list->head->prev;
		mmcp_ti_unlink(entry);
		list->num--;
	}
	else {
		/* If alarm is head, ... */
		if( alarm == list->head )
			list->head = alarm->next;

		/* If alarm is tail, ... */
		else if( alarm == list->tail )
			list->tail = alarm->prev;

		entry = alarm;
		mmcp_ti_unlink(entry);
		list->num--;
	}

	/* Although the timer list is empty, 'head'&'tail' of 'list' point 'entry'.
	 * Therefore we must modify them to NULL. */
	if( mmcp_timerlist_empty(list) )
		list->head = list->tail = NULL;

	if( mmcp_mutex_off(&list->mutex) < 0 )	// check
		err_ret( "mmcp_timerlist_del() : mmcp_mutex_off() failed.", entry );

	return entry;
}

/* ----------------------------------------------------------------- *
 *	Implementation of Timer Dispatcher
 * ----------------------------------------------------------------- */

void mmcp_timer_dispatch( struct mmcp_socket *mso )
{
	struct mmcp_endpoint *ep = mso->ep;
	
	struct timeval *to = NULL;
	struct mmcp_timeritem *timer = NULL;
	struct mmcp_skbuff *skb = NULL;
//	struct mmcp_basehdr *hdr = NULL;

	struct mmcp_eventitem *e = NULL;

	/* If the timer list is empty, just return. */
	if( !(to = mmcp_timerlist_peek_timeout(&mso->timer_list)) )
		return;

	/* If not expired, just return. */
	if( to->tv_sec != 0 || to->tv_usec != 0 )
		return;

	/* If expired, ... */
	timer = mmcp_timerlist_del( &mso->timer_list, NULL );
	if( !timer )
		return;

	if( timer->type != MMCP_TIMER_SRT && timer->type != MMCP_TIMER_SPT ) {
		skb = timer->skb;
//		hdr = (struct mmcp_basehdr *) skb->pkt;
	}

	switch( timer->type ) {

		case MMCP_TIMER_SRT:

			if( ep->role == MMCP_ROLE_MN ) {
				e = mmcp_create_event( MMCP_SND_STATUS_REPORT, 0,
							NULL, 0, NULL, 0 );
			}

			else if( ep->role == MMCP_ROLE_LMC ) {
				e = mmcp_create_event( MMCP_SND_ASTATUS_REPORT, 0,
							NULL, 0, NULL, 0 );
			}

			else
				err_msg( "mmcp_timer_dispatch() : SRT is used by MN, LMC" );

			if( !e ) {
				err_msg( "mmcp_timer_dispatch() : "
						 "mmcp_create_event() failed." );
				break;
			}

			if( mmcp_event_enq(&mso->eventq, e) < 0 ) {
				err_msg( "mmcp_timer_dispatch() : "
						 "mmcp_event_enq() failed." );
				mmcp_destroy_event(e);
			}

			break;

		case MMCP_TIMER_SPT:

			e = mmcp_create_event( MMCP_SND_STATUS_PROBE, 0,
									NULL, 0, NULL, 0 );

			if( !e ) {
				err_msg( "mmcp_timer_dispatch() : "
						 "mmcp_create_event() failed." );
				break;
			}

			if( mmcp_event_enq(&mso->eventq, e) < 0 ) {
				err_msg( "mmcp_timer_dispatch() : "
						 "mmcp_create_enq() failed." );
				mmcp_destroy_event(e);
				break;
			}

			break;

		case MMCP_TIMER_JWT:
		case MMCP_TIMER_HIT:
		case MMCP_TIMER_CTT:
			if( ++skb->rtxcnt == MMCP_MAX_RXT ) {
				e = mmcp_create_event( MMCP_EXP_HANDOVER_TRANSFER, 0,
						&skb->addr, skb->addrlen, NULL, 0 );

				if( !e ) {
					err_msg( "mmcp_timer_dispatch() : "
							 "mmcp_create_event() failed." );
					break;
				}

				if( mmcp_event_enq(&mso->handoverq, e) < 0 ) {
					err_msg( "mmcp_timer_dispatch() : "
							 "mmcp_handover_enq() failed." );
					mmcp_destroy_event(e);
				}

				break;

			}
		case MMCP_TIMER_RXT:
			/*	Check if the retransmission counter does exceed maximum
			 *	maximum retransmit number that is the same value 
			 *	as MMCP_MAX_RTX
			 */
			if( ++skb->rtxcnt == MMCP_MAX_RXT ) {

				switch( ep->state ) {
					case MMCP_STATE_SJR_SENT:
						e = mmcp_create_event( MMCP_EXP_SESSION_JOIN, 0,
								&skb->addr, skb->addrlen, NULL, 0 );

						if( !e ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_create_event() failed." );
							break;
						}

						if( mmcp_event_enq(&mso->joinq, e) < 0 ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_event_enq() failed." );
							mmcp_destroy_event(e);
						}

						break;

					case MMCP_STATE_LJR_SENT:
						e = mmcp_create_event( MMCP_EXP_LOCAL_JOIN, 0,
								&skb->addr, skb->addrlen, NULL, 0 );

						if( !e ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_create_event() failed." );
							break;
						}

						if( mmcp_event_enq(&mso->joinq, e) < 0 ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_event_enq() failed." );
							mmcp_destroy_event(e);
						}

						break;

					case MMCP_STATE_HR_SENT:
						e = mmcp_create_event( MMCP_EXP_HANDOVER_INIT, 0,
								&skb->addr, skb->addrlen, NULL, 0 );

						if( !e ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_create_event() failed." );
							break;
						}

						if( mmcp_event_enq(&mso->handoverq, e) < 0 ) {
							err_msg( "mmcp_timer_dispatch() : "
									 "mmcp_event_enq() failed." );
							mmcp_destroy_event(e);
						}

						break;

					default:
						break;
				}	/* switch( ep->state ) */
			}	/* if( ++skb->rtxcnt == MMCP_MAX_RTX ) */

			/* Recalulate the action timer of 'timer' and
			 * move it to the end of the its list.
			 */

			if( mmcp_ti_update(timer) < 0 )
				err_goto( "mmcp_timer_dispatch() : mmcp_ti_update() failed.",
							discard_timer );



			if( mmcp_timerlist_add(timer->list, timer) < 0 )
				err_goto( "mmcp_timer_dispatch() : "
							"mmcp_timerlist_add() failed.", 
							discard_timer );

		//	if( mmcp_skb_retransmit(skb) < 0 )
			if( mmcp_skb_transmit(skb) < 0 )
				err_msg( "mmcp_timer_dispatch() : "
						 "mmcp_skb_transmit() failed." );

			break;

		default:
			break;
	}	/* switch( timer->type ) */

	return;

discard_timer:
	mmcp_ti_free(timer);
	return;
}
