/*	Mobile Multicast Control Protocol Implementation.
 *
 *	This file is part of the MMCP over UDP Implementation.
 *
 *	user.c
 *	This file includes the implementation of the MMCP user API.
 *
 *	Contact :	Jaesung Park(knucsid@gmail.com)
 *				Jiin Kim(jiin16@gmail.com)
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <header/mmcp.h>
#include <header/types.h>
#include <header/structs.h>
#include <header/user.h>
#include <header/wrapper.h>


/* ----------------------------------------------------------------- *
 *	User APIs
 * ----------------------------------------------------------------- */

/* user.h ---------------------------------------------------------- */

int mmcp_socket( int domain, int protocol, int role )
{
	struct mmcp_socket *mso = NULL;

	mso = mmcp_create_socket( domain, protocol, (mmcp_role_t) role );
	if( !mso ) {
		err_ret( "mmcp_socket() : mmcp_create_socket() failed.", -1 );
	}

	if( mmcp_create_endpoint( mso, (mmcp_role_t) role ) < 0 ) {
		mmcp_destroy_socket( mso );
		err_ret( "mmcp_socket() : mmcp_create_endpoint() failed.", -1 );
	}

	/* Socket created! (STATE: CLOSED) */
	fputs("MMCP, Socket Created! (STATE: CLOSED)\n", stdout);

	return mso->msd;
}

int mmcp_create( int msd, struct sockaddr_storage *local_addr,
	   				struct sockaddr_storage *group_addr	)
{
	struct mmcp_socket *mso = NULL;
	struct mmcp_sm *sm = NULL;

	/* Get current socket for the given 'msd' */
	mso = mmcp_get_socket(msd);
	if( !mso )
		err_ret( "mmcp_create() : invalid socket descriptor.", -1 );

	/* Check The Role */
	if( mso->ep->role == MMCP_ROLE_SM ) {

		sm = mso->ep->sm;

		if( mmcp_sk_bind(mso, local_addr) < 0 )
			err_ret( "mmcp_create() : mmcp_sk_bind() failed.", -1 );

		memcpy( &sm->group_addr, group_addr, mmcp_addrlen(mso->domain) );

		if( mmcp_ops_start_events(mso) < 0 )
			err_ret( "mmcp_create() : mmcp_ops_start_events() failed.", -1 );
	}

	else {
		err_ret( "mmcp_create() : Only SM can use this function.", -1 );
	}

	return 0;
}

//int mmcp_join( int msd, int usd, struct sockaddr_storage *local_addr,
int mmcp_join( int msd, int usd, 
	struct sockaddr_storage *upper_addr, struct sockaddr_storage *mcs_addr )
{
	struct mmcp_socket *mso = NULL;
	int result = -1;
	int addrlen;
	struct timeval init_time;

	/* Get Current Socket for The Given 'msd' */
	mso = mmcp_get_socket(msd);
	if( !mso )
		err_ret( "mmcp_join() : invalid socket descriptor.", -1 );

	/* Check The Role */
	if( mso->ep->role != MMCP_ROLE_MN )
		err_ret( "mmcp_join() : Join must be MN.", -1 );
	
	else if( mso->ep->role == MMCP_ROLE_MN ) {
		/* Socket Descriptor for Multicast Data */
		mso->usd = usd;
		addrlen = mmcp_addrlen( mso->domain );
		memcpy( &mso->ep->mn->sm_addr, upper_addr, addrlen );
	}

	else {
		err_ret( "mmcp_join() : Role error.", -1 );
	}


	/* Invoke Main Loop and Event Handlers. */
	if( mmcp_ops_start_events(mso) < 0 )
		err_ret( "mmcp_join() : mmcp_ops_start_event() failed.", -1 );

	result = mmcp_ops_start_join( mso );
	
	addrlen = mmcp_addrlen( mso->domain );
	memcpy( mcs_addr, &mso->ep->mn->group_addr, addrlen );

	gettimeofday( &init_time, NULL );
	mso->ep->mn->elapsed_time = init_time.tv_sec;

	return result;
}

int mmcp_lmc_join( int msd, struct sockaddr_storage *local_addr,
				struct sockaddr_storage *upper_addr )
{
	struct mmcp_socket *mso = NULL;
	int addrlen;

	/* Get Current Socket for The Given 'msd' */
	mso = mmcp_get_socket(msd);
	if( !mso )
		err_ret( "mmcp_lmc_join() : invalid socket descriptor.", -1 );

	/* Check The Role */
	if( mso->ep->role != MMCP_ROLE_LMC )
		err_ret( "mmcp_lmc_join() : Join must be either LMC or MN.", -1 );
	
	else if( mso->ep->role == MMCP_ROLE_LMC ) {
		addrlen = mmcp_addrlen( mso->domain );
		memcpy( &mso->ep->lmc->sm_addr, upper_addr, addrlen );
	}

	else {
		err_ret( "mmcp_lmc_join() : Role error.", -1 );
	}


	if( mmcp_sk_bind(mso, local_addr) < 0 )
		err_ret( "mmcp_lmc_join() : mmcp_sk_bind() failed.", -1 );

	if( mso->addr.sa.sin_port == htons(MMCP_LMC_PORT_ONE) )
		memcpy( mso->ep->lmc->poa_id, MMCP_POA_ONE, sizeof(MMCP_POA_ONE) );
	else
		memcpy( mso->ep->lmc->poa_id, MMCP_POA_TWO, sizeof(MMCP_POA_TWO) );

	/* Invoke Main Loop and Event Handlers. */
	if( mmcp_ops_start_events(mso) < 0 )
		err_ret( "mmcp_lmc_join() : mmcp_ops_start_event() failed.", -1 );

	if( mmcp_ops_start_join(mso) < 0 )
		err_ret( "mmcp_lmc_join() : mmcp_ops_start_join() failed.", -1 );

	if( mmcp_ops_maintain_events(mso) != 0 )
		err_ret( "mmcp_lmc_join() : mmcp_ops_maintain_event() failed.", -1 );
	

	return 0;
}

void mmcp_processing( int msd )
{
	struct mmcp_socket *mso = NULL;

	/* Get current socket for the given 'msd' */
	mso = mmcp_get_socket(msd);
	if( !mso )
		err_msg( "mmcp_processing() : invalid socket descriptor." );

	struct mmcp_handleset *hset = &mso->hset;

	struct timeval *to;
	struct timeval default_to;

	int result;
	fd_set backup_fds, read_fds;
	FD_ZERO(&backup_fds);
	FD_SET( mso->msd, &backup_fds );

	mso->ep->state = MMCP_STATE_ESTABLISHED;

	puts( "MMCP, Creation Completed! (STATE: ESTABLISHED)" );

	while( hset->main_loop_is_running ) {
		read_fds = backup_fds;

		if( !(to = mmcp_timerlist_peek_timeout(&mso->timer_list)) ) {
			default_to.tv_sec	= MMCP_MAINLOOP_TIMEOUT_SEC;
			default_to.tv_usec	= MMCP_MAINLOOP_TIMEOUT_USEC;
			to = &default_to;
		}

		result = select( mso->msd + 1, &read_fds, NULL, NULL, to );

		switch( result ) {
			case -1:
				result = 0;
				err_msg( "mmcp_processing() : select() failed." );
				break;
			case 0:
				mmcp_timer_dispatch(mso);
				break;
			default:
				mmcp_event_dispatch( mso, &read_fds );
				break;
		}
	}
}

int mmcp_leave( int msd )
{
	return 0;
}


int mmcp_setsockopt( int msd, int level, int optname,
							void *optval, socklen_t optlen )
{
	return 0;
}

int mmcp_getsockopt( int msd, int level, int optname,
							void *optval, socklen_t *optlen )
{
	return 0;
}

int mmcp_status_monitor()
{
	return 0;
}

int mmcp_test_handover( int msd )
{
	struct mmcp_socket *mso = NULL;

	/* Get current socket for the given 'msd' */
	mso = mmcp_get_socket(msd);
	if( !mso )
		err_ret( "mmcp_test_handover() : invalid socket descriptor.", -1 );

	if( mmcp_ops_request_handover(mso) < 0 ) {
		err_ret( "mmcp_test_handover() : "
				 "mmcp_ops_request_handover() failed.", -1 );
	}

	if( mmcp_ops_local_join(mso) < 0 ) {
		err_ret( "mmcp_test_handover() : "
				 "mmcp_ops_request_handover() failed.", -1 );
	}
	return 0;
}
